import datetime
from typing import List

from .kawa_types import Types


def inputs(**kwargs):
    def decorator_set_inputs(func):
        func.inputs = [{'name': k, 'type': python_type_to_kawa_type(v)} for k, v in kwargs.items()]
        return func

    return decorator_set_inputs


def outputs(**kwargs):
    def decorator_set_outputs(func):
        func.outputs = [{'name': k, 'type': python_type_to_kawa_type(v)} for k, v in kwargs.items()]
        return func

    return decorator_set_outputs


def secrets(**kwargs):
    def decorator_set_secret_mapping(func):
        func.secrets = kwargs
        return func

    return decorator_set_secret_mapping


def kawa_tool(inputs: dict = None, outputs: dict = None, secrets: dict = None):
    def decorator(func):
        _in = inputs or {}
        _out = outputs or {}
        func.inputs = [{'name': k, 'type': python_type_to_kawa_type(v)} for k, v in _in.items()]
        func.outputs = [{'name': k, 'type': python_type_to_kawa_type(v)} for k, v in _out.items()]
        func.secrets = secrets or {}
        return func

    return decorator


def python_type_to_kawa_type(python_type):
    if python_type == str:
        return Types.TEXT
    if python_type == float:
        return Types.DECIMAL
    if python_type == datetime.date:
        return Types.DATE
    if python_type == datetime.datetime:
        return Types.DATE_TIME
    if python_type == bool:
        return Types.BOOLEAN
    if python_type == List[float]:
        return Types.LIST_OF_DECIMALS

    raise Exception('This type is not yet available: {}'.format(python_type))



import pandas as pd
from datetime import date, datetime
import uuid

from ..kawa_base_e2e_test import KawaBaseTest
from ...client.kawa_client import KawaClient as kawa
import zoneinfo


class TestPartitionManagement(KawaBaseTest):

    def test_that_data_is_deleted_based_on_a_date_time(self):
        # setup
        eastern_tz = zoneinfo.ZoneInfo('US/Eastern')
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {
                'trade_date': datetime(2024, 1, 1, 10, 0, 0, tzinfo=eastern_tz),
                'ticker': 'AAPL',
                'price': 112.3
            },
            {
                'trade_date': datetime(2024, 1, 1, 11, 0, 0, tzinfo=eastern_tz),
                'ticker': 'AAPL',
                'price': 113.3
            },
            {
                'trade_date': datetime(2024, 1, 1, 12, 0, 0, tzinfo=eastern_tz),
                'ticker': 'AAPL',
                'price': 114.3
            },
            {
                'trade_date': datetime(2024, 1, 1, 13, 0, 0, tzinfo=eastern_tz),
                'ticker': 'AAPL',
                'price': 115.3
            },
            {
                'trade_date': datetime(2024, 1, 1, 10, 0, 0, tzinfo=eastern_tz),
                'ticker': 'TSLA',
                'price': 12.3
            },
            {
                'trade_date': datetime(2024, 1, 1, 11, 0, 0, tzinfo=eastern_tz),
                'ticker': 'TSLA',
                'price': 13.3
            },
            {
                'trade_date': datetime(2024, 1, 1, 12, 0, 0, tzinfo=eastern_tz),
                'ticker': 'TSLA',
                'price': 42.3
            },
            {
                'trade_date': datetime(2024, 1, 1, 13, 0, 0, tzinfo=eastern_tz),
                'ticker': 'TSLA',
                'price': 32.3
            },
        ])

        loader = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader.create_datasource()
        datasource = loader.load_data(reset_before_insert=True, create_sheet=True)

        # given
        delete_from = datetime(2024, 1, 1, 12, 0, 0, tzinfo=eastern_tz)
        where_clauses = [
            # Will delete everything that is after or on delete_from
            kawa.where('trade_date').datetime_range(from_inclusive=delete_from)
        ]

        # when
        self.commands.delete_data(datasource=datasource, delete_where=where_clauses)
        computed_df_after_data_delete = (self.kawa.sheet(sheet_name=unique_id)
                                         .select('trade_date', 'ticker', 'price')
                                         .compute())

        # then
        self.assertListEqual(list(computed_df_after_data_delete['trade_date']), [
            datetime(2024, 1, 1, 10, 0, 0, tzinfo=eastern_tz),
            datetime(2024, 1, 1, 11, 0, 0, tzinfo=eastern_tz),
            datetime(2024, 1, 1, 10, 0, 0, tzinfo=eastern_tz),
            datetime(2024, 1, 1, 11, 0, 0, tzinfo=eastern_tz)
        ])
        self.assertListEqual(list(computed_df_after_data_delete['ticker']), ['AAPL', 'AAPL', 'TSLA', 'TSLA'])
        self.assertListEqual(list(computed_df_after_data_delete['price']), [112.3, 113.3, 12.3, 13.3])

    def test_that_data_is_deleted_based_on_a_date_time_with_different_time_zone(self):
        # setup
        eastern_tz = zoneinfo.ZoneInfo('US/Eastern')
        pacific_tz = zoneinfo.ZoneInfo('US/Pacific')

        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {
                'trade_date': datetime(2024, 1, 1, 10, 0, 0, tzinfo=eastern_tz),
                'ticker': 'AAPL',
                'price': 112.3
            },
            {
                'trade_date': datetime(2024, 1, 1, 11, 0, 0, tzinfo=eastern_tz),
                'ticker': 'AAPL',
                'price': 113.3
            },
            {
                'trade_date': datetime(2024, 1, 1, 12, 0, 0, tzinfo=eastern_tz),
                'ticker': 'AAPL',
                'price': 114.3
            },
            {
                'trade_date': datetime(2024, 1, 1, 13, 0, 0, tzinfo=eastern_tz),
                'ticker': 'AAPL',
                'price': 115.3
            },
            {
                'trade_date': datetime(2024, 1, 1, 10, 0, 0, tzinfo=eastern_tz),
                'ticker': 'TSLA',
                'price': 12.3
            },
            {
                'trade_date': datetime(2024, 1, 1, 11, 0, 0, tzinfo=eastern_tz),
                'ticker': 'TSLA',
                'price': 13.3
            },
            {
                'trade_date': datetime(2024, 1, 1, 12, 0, 0, tzinfo=eastern_tz),
                'ticker': 'TSLA',
                'price': 42.3
            },
            {
                'trade_date': datetime(2024, 1, 1, 13, 0, 0, tzinfo=eastern_tz),
                'ticker': 'TSLA',
                'price': 32.3
            },
        ])

        loader = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader.create_datasource()
        datasource = loader.load_data(reset_before_insert=True, create_sheet=True)

        # given
        # 10:00 Pacific is 13:00 Eastern
        delete_from = datetime(2024, 1, 1, 10, 0, 0, tzinfo=pacific_tz)
        where_clauses = [
            # Will delete everything that is after or on delete_from
            kawa.where('trade_date').datetime_range(from_inclusive=delete_from)
        ]

        # when
        self.commands.delete_data(datasource=datasource, delete_where=where_clauses)
        computed_df_after_data_delete = (self.kawa.sheet(sheet_name=unique_id)
                                         .select('trade_date', 'ticker', 'price')
                                         .compute())

        # then
        self.assertListEqual(list(computed_df_after_data_delete['trade_date']), [
            datetime(2024, 1, 1, 10, 0, 0, tzinfo=eastern_tz),
            datetime(2024, 1, 1, 11, 0, 0, tzinfo=eastern_tz),
            datetime(2024, 1, 1, 12, 0, 0, tzinfo=eastern_tz),
            datetime(2024, 1, 1, 10, 0, 0, tzinfo=eastern_tz),
            datetime(2024, 1, 1, 11, 0, 0, tzinfo=eastern_tz),
            datetime(2024, 1, 1, 12, 0, 0, tzinfo=eastern_tz)
        ])
        self.assertListEqual(list(computed_df_after_data_delete['ticker']),
                             ['AAPL', 'AAPL', 'AAPL', 'TSLA', 'TSLA', 'TSLA'])
        self.assertListEqual(list(computed_df_after_data_delete['price']), [112.3, 113.3, 114.3, 12.3, 13.3, 42.3])

    def test_that_data_is_deleted_based_on_a_number_filter(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {'date': date(2024, 1, 1), 'int': 1, 'other': 100},
            {'date': date(2024, 1, 1), 'int': 2, 'other': 10},
            {'date': date(2024, 1, 1), 'int': 3, 'other': 10},
            {'date': date(2024, 1, 2), 'int': 4, 'other': 10},
            {'date': date(2024, 1, 2), 'int': 5, 'other': 10},
            {'date': date(2024, 1, 3), 'int': 6, 'other': 10},
            {'date': date(2024, 1, 3), 'int': 7, 'other': 10},
        ])

        loader = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader.create_datasource()
        datasource = loader.load_data(reset_before_insert=True, create_sheet=True)
        self.kawa.entities.get_datasource_schema(datasource_id=datasource.get('id')).get('schema')

        # given
        computed_df_before_delete = self.kawa.sheet(sheet_name=unique_id).select('date', 'int', 'other').compute()
        where_clauses = [
            kawa.where('int').lte(4),
            kawa.where('other').ne(10).exclude(),
        ]

        # when
        self.commands.delete_data(datasource=datasource, delete_where=where_clauses)
        computed_df_after_data_delete = self.kawa.sheet(sheet_name=unique_id).select('date', 'int', 'other').compute()

        # then: computation before drop is correct
        self.assertListEqual(list(computed_df_before_delete['int']), list(range(1, 8)))
        self.assertListEqual(list(computed_df_before_delete['other']), [100, 10, 10, 10, 10, 10, 10])
        self.assertListEqual(list(computed_df_before_delete['date']), [
            date(2024, 1, 1),
            date(2024, 1, 1),
            date(2024, 1, 1),
            date(2024, 1, 2),
            date(2024, 1, 2),
            date(2024, 1, 3),
            date(2024, 1, 3)
        ])

        # and: computation after drop is correct
        self.assertListEqual(list(computed_df_after_data_delete['int']), [1, 5, 6, 7])
        self.assertListEqual(list(computed_df_after_data_delete['other']), [100, 10, 10, 10])
        self.assertListEqual(list(computed_df_after_data_delete['date']), [
            date(2024, 1, 1),
            date(2024, 1, 2),
            date(2024, 1, 3),
            date(2024, 1, 3)
        ])

    def test_that_data_is_deleted_based_on_a_text_filter(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {'ticker': 'AAPL'},
            {'ticker': 'TSLA'},
            {'ticker': 'MSFT'},
            {'ticker': 'AMZN'},
        ])

        loader = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader.create_datasource()
        datasource = loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        where_clauses = [kawa.where('ticker').in_list(['TSLA', 'MSFT'])]
        self.commands.delete_data(datasource=datasource, delete_where=where_clauses)
        computed_df_after_data_delete = self.kawa.sheet(sheet_name=unique_id).select('ticker').compute()

        # then: computation after drop is correct
        self.assertListEqual(list(computed_df_after_data_delete['ticker']), ['AAPL', 'AMZN'])

    def test_that_data_is_deleted_based_on_a_combination_of_text_filters(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {'ticker': 'AAPL'},
            {'ticker': 'TSLA'},
            {'ticker': 'MSFT'},
            {'ticker': 'AMZN'},
        ])

        loader = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader.create_datasource()
        datasource = loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        where_clauses = [
            kawa.where('ticker').contains('M'),
            kawa.where('ticker').contains('N')
        ]
        self.commands.delete_data(datasource=datasource, delete_where=where_clauses)
        computed_df_after_data_delete = self.kawa.sheet(sheet_name=unique_id).select('ticker').compute()

        # then: computation after drop is correct
        self.assertListEqual(list(computed_df_after_data_delete['ticker']), ['AAPL', 'TSLA', 'MSFT'])

    def test_that_data_is_deleted_based_on_a_date_filter(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {'date': date(2024, 1, 1)},
            {'date': date(2024, 1, 2)},
            {'date': date(2024, 1, 3)},
            {'date': date(2024, 1, 4)},
        ])

        loader = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader.create_datasource()
        datasource = loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        where_clauses = [
            kawa.where('date').date_range(
                from_inclusive=date(2024, 1, 2),
                to_inclusive=date(2024, 1, 3))
        ]
        self.commands.delete_data(datasource=datasource, delete_where=where_clauses)
        computed_df_after_data_delete = self.kawa.sheet(sheet_name=unique_id).select('date').compute()

        # then: computation after drop is correct
        self.assertListEqual(list(computed_df_after_data_delete['date']), [
            date(2024, 1, 1),
            date(2024, 1, 4)
        ])

    def test_that_data_is_deleted_based_on_a_combination_of_date_time_and_text_filters(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        dtf1 = pd.DataFrame([
            {'k': 1, 'ticker': 'AAPL', 'trade_date': datetime(2024, 1, 1, 13, 0, 0)},
            {'k': 2, 'ticker': 'AAPL', 'trade_date': datetime(2024, 1, 1, 13, 1, 0)},
            {'k': 3, 'ticker': 'AAPL', 'trade_date': datetime(2024, 1, 1, 13, 2, 0)},
            {'k': 4, 'ticker': 'AAPL', 'trade_date': datetime(2024, 1, 1, 13, 3, 0)},
            {'k': 5, 'ticker': 'MSFT', 'trade_date': datetime(2024, 1, 1, 13, 0, 0)},
            {'k': 6, 'ticker': 'MSFT', 'trade_date': datetime(2024, 1, 1, 13, 1, 0)},
            {'k': 7, 'ticker': 'MSFT', 'trade_date': datetime(2024, 1, 1, 13, 2, 0)},
            {'k': 8, 'ticker': 'MSFT', 'trade_date': datetime(2024, 1, 1, 13, 3, 0)},
        ])

        loader = self.kawa.new_data_loader(df=dtf1, datasource_name=unique_id)
        loader.create_datasource()
        datasource = loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        utc = zoneinfo.ZoneInfo('UTC')
        from_date = datetime(2024, 1, 1, 13, 1, 0, tzinfo=utc)
        where_clauses = [
            kawa.where('ticker').in_list(['MSFT']),
            kawa.where('trade_date').datetime_range(from_inclusive=from_date)
        ]
        self.commands.delete_data(datasource=datasource, delete_where=where_clauses)
        computed_df_after_data_delete = self.kawa.sheet(sheet_name=unique_id).select('k').compute()

        # then: computation after drop is correct
        self.assertListEqual(list(computed_df_after_data_delete['k']), [1, 2, 3, 4, 5])



import datetime
from dataclasses import dataclass
from os import path

import pandas as pd
import pandas.testing as pd_testing
import uuid
import time

import pytz
from numpy import nan

from ..kawa_base_e2e_test import KawaBaseTest


class TestPythonColumns(KawaBaseTest):

    def test_that_python_column_is_correctly_computed_case_with_no_pk_in_param_of_script(self):
        pk_name = 'pk strange name%%1'
        df_for_datasource = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22}]
        )
        indicator_for_param_mapping = ['measure1', 'measure2']
        expected_df = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2, 'measure3': 3.3},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22, 'measure3': 33.33}]
        )
        self.run_test_python_column_execution(df_for_datasource,
                                              [pk_name],
                                              'Columns with no pk in params',
                                              indicator_for_param_mapping,
                                              expected_df)

    def test_that_python_column_is_correctly_computed_case_with_no_pk_in_param_of_script_with_a_date_pk(self):
        df_for_datasource = pd.DataFrame([
            {'pk': datetime.date(2024, 1, 1), 'measure1': 1.1, 'measure2': 2.2},
            {'pk': datetime.date(2024, 1, 2), 'measure1': 11.11, 'measure2': 22.22}]
        )
        indicator_for_param_mapping = ['measure1', 'measure2']
        expected_df = pd.DataFrame([
            {'pk': datetime.date(2024, 1, 1), 'measure1': 1.1, 'measure2': 2.2, 'measure3': 3.3},
            {'pk': datetime.date(2024, 1, 2), 'measure1': 11.11, 'measure2': 22.22, 'measure3': 33.33}]
        )
        self.run_test_python_column_execution(df_for_datasource,
                                              ['pk'],
                                              'Columns with no pk in params',
                                              indicator_for_param_mapping,
                                              expected_df)

    def test_that_python_column_is_correctly_computed_case_with_no_pk_in_param_of_script_with_a_datetime_pk(self):
        df_for_datasource = pd.DataFrame([
            {'pk': at_berlin(datetime.datetime(2024, 1, 1, 1)), 'measure1': 1.1, 'measure2': 2.2},
            {'pk': at_berlin(datetime.datetime(2024, 1, 1, 2)), 'measure1': 11.11, 'measure2': 22.22}]
        )
        indicator_for_param_mapping = ['measure1', 'measure2']
        expected_df = pd.DataFrame([
            {'pk': at_berlin(datetime.datetime(2024, 1, 1, 1)), 'measure1': 1.1, 'measure2': 2.2, 'measure3': 3.3},
            {'pk': at_berlin(datetime.datetime(2024, 1, 1, 2)), 'measure1': 11.11, 'measure2': 22.22,
             'measure3': 33.33}]
        )
        self.run_test_python_column_execution(df_for_datasource,
                                              ['pk'],
                                              'Columns with no pk in params',
                                              indicator_for_param_mapping,
                                              expected_df)

    def test_that_python_column_using_all_types(self):
        df_for_datasource = pd.DataFrame([
            all_type_row(0, 1.1, 1, datetime.date(2024, 7, 14), datetime.datetime(2024, 7, 14, 14, 7, 0), True, 'true'),
            all_type_row(1, 2.2, 2, datetime.date(2024, 7, 15), datetime.datetime(2024, 7, 15, 14, 7, 0), False,
                         'false')
        ]
        )
        row1 = all_type_row(0, 1.1, 1, datetime.date(2024, 7, 14), datetime.datetime(2024, 7, 14, 14, 7, 0), True,
                            'true')
        row1.update(
            all_type_row_res(2.2, 2, datetime.date(2024, 7, 15),
                             datetime.datetime(2024, 7, 14, 15, 7, 0), False,
                             'True__T__'))
        row2 = all_type_row(1, 2.2, 2, datetime.date(2024, 7, 15), datetime.datetime(2024, 7, 15, 14, 7, 0), False,
                            'false')
        row2.update(
            all_type_row_res(3.3, 3, datetime.date(2024, 7, 16),
                             datetime.datetime(2024, 7, 15, 15, 7, 0), True,
                             'False__T__'))
        indicator_for_param_mapping = ['decimal', 'integer', 'date', 'date_time', 'boolean', 'text']
        expected_df = pd.DataFrame([row1, row2])
        self.run_test_python_column_execution(df_for_datasource,
                                              ['pk'],
                                              'Column script with all types',
                                              indicator_for_param_mapping,
                                              expected_df,
                                              secrets={'secret_name_in_kawa': '__T__'})

    def test_that_python_column_is_correctly_computed_case_with_pk_in_param_of_script(self):
        pk_name = 'pk strange name%%1'
        df_for_datasource = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22}]
        )
        indicator_for_param_mapping = [pk_name, 'measure1', 'measure2']
        expected_df = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2, 'measure3': 3.3},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22, 'measure3': 33.33}]
        )
        self.run_test_python_column_execution(df_for_datasource,
                                              [pk_name],
                                              'Column script with pk in params',
                                              indicator_for_param_mapping,
                                              expected_df)

    def test_that_logs_are_accessible_for_the_python_columns(self):
        df_for_datasource = pd.DataFrame([
            {'pk': 1, 'measure1': 1.1, 'measure2': 2.2},
            {'pk': 2, 'measure1': 11.11, 'measure2': 22.22}]
        )
        indicator_for_param_mapping = ['measure1', 'measure2']
        expected_df = pd.DataFrame([
            {'pk': 1, 'measure1': 1.1, 'measure2': 2.2, 'measure3': 3.3},
            {'pk': 2, 'measure1': 11.11, 'measure2': 22.22, 'measure3': 33.33}]
        )
        self.run_test_python_column_execution(df_for_datasource,
                                              ['pk'],
                                              'Column script with no pk in params with logging',
                                              indicator_for_param_mapping,
                                              expected_df,
                                              logs='Computing the df hahahahahahahahah')

    def test_that_python_column_is_correctly_computed_case_with_two_pk__only_one_in_param_of_script(self):
        df_for_datasource = pd.DataFrame([
            {'pk': 1, 'pk2': 0, 'measure1': 1.1, 'measure2': 2.2},
            {'pk': 2, 'pk2': 1, 'measure1': 11.11, 'measure2': 22.22}]
        )
        indicator_for_param_mapping = ['pk', 'measure1', 'measure2']
        expected_df = pd.DataFrame([
            {'pk': 1, 'pk2': 0, 'measure1': 1.1, 'measure2': 2.2, 'measure3': 3.3},
            {'pk': 2, 'pk2': 1, 'measure1': 11.11, 'measure2': 22.22, 'measure3': 33.33}]
        )
        self.run_test_python_column_execution(df_for_datasource,
                                              ['pk', 'pk2'],
                                              'Column script with pk in params',
                                              indicator_for_param_mapping,
                                              expected_df)

    def test_multi_user_python_column_usage(self):
        # First user 1 create DS, sheet, layout, script and compute it.
        df_for_datasource = pd.DataFrame([
            {'pk': 1, 'measure1': 1.1, 'measure2': 2.2},
            {'pk': 2, 'measure1': 11.11, 'measure2': 22.22}]
        )
        indicator_for_param_mapping = ['measure1', 'measure2']
        expected_df = pd.DataFrame([
            {'pk': 1, 'measure1': 1.1, 'measure2': 2.2, 'measure3': 3.3},
            {'pk': 2, 'measure1': 11.11, 'measure2': 22.22, 'measure3': 33.33}]
        )
        test_result = self.run_test_python_column_execution(df_for_datasource,
                                                            ['pk'],
                                                            'Columns with no pk in params',
                                                            indicator_for_param_mapping,
                                                            expected_df)

        # Now user 2 computes the shared layout, he should see the 'measure3' column, but with no values
        second_user_computation_res = self.kawa_second_user.sheet().view_id(test_result.default_layout_id).compute()
        expected_df_second_user = pd.DataFrame([
            {'pk': 1, 'measure1': 1.1, 'measure2': 2.2, 'measure3': nan},
            {'pk': 2, 'measure1': 11.11, 'measure2': 22.22, 'measure3': nan}]
        )
        pd_testing.assert_frame_equal(second_user_computation_res, expected_df_second_user)

        # When: User 2 launches a script compute
        self.commands_second_user.start_python_private_join(test_result.python_private_join['id'])
        time.sleep(10)
        second_user_computation_res_after_script_execution = self.kawa_second_user.sheet() \
            .view_id(test_result.default_layout_id).compute()

        # Then user 2 should see the expected df with the 'measure3' computed values
        pd_testing.assert_frame_equal(second_user_computation_res_after_script_execution, expected_df)

    def test_script_execution_with_secret(self):
        df_for_datasource = pd.DataFrame([
            {'pk': 1, 'measure1': 1.1, 'measure2': 2.2},
            {'pk': 2, 'measure1': 11.11, 'measure2': 22.22}]
        )
        indicator_for_param_mapping = ['measure1']
        expected_df = pd.DataFrame([
            {'pk': 1, 'measure1': 1.1, 'measure2': 2.2, 'measure3': 11.1},
            {'pk': 2, 'measure1': 11.11, 'measure2': 22.22, 'measure3': 21.11}]
        )

        self.run_test_python_column_execution(df_for_datasource,
                                              ['pk'],
                                              'Column script with secrets',
                                              indicator_for_param_mapping,
                                              expected_df,
                                              {'delta_in_letters': '10_letters'})

    def test_python_column_execution_with_update_of_script_to_add_outputs_and_then_execute(self):
        pk_name = 'pk strange name%%1'
        df_for_datasource = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22}]
        )
        indicator_for_param_mapping = ['measure1', 'measure2']
        expected_df = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2, 'measure3': 3.3},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22, 'measure3': 33.33}]
        )
        res = self.run_test_python_column_execution(df_for_datasource,
                                                    [pk_name],
                                                    'Columns with no pk in params',
                                                    indicator_for_param_mapping,
                                                    expected_df)

        script_id = res.python_private_join['scriptId']

        script = self.commands.replace_script(script_id,
                                              'my amazing new name',
                                              'kywy-specs-final',
                                              'Columns with no pk in params with new output')

        script_outputs = script['scriptMetadata']['outputs']

        # check output has been added to script metadata
        self.assertTrue(len(script_outputs) == 2)

        # now start the python private join
        self.commands.start_python_private_join(res.python_private_join['id'])

        # wait for the computation of the script ot finish
        time.sleep(10)
        # check the indicators in the python data source
        data_source_id = res.python_private_join['joinDatasourceId']
        data_source = self.kawa.entities.datasources().get_entity_by_id(data_source_id)
        indicator_ids = [i['indicatorId'] for i in data_source['indicators']]
        self.assertTrue('measure4' in indicator_ids)

        # add the new indicator to the layout to check computation is ok
        sheet_with_private_join = self.kawa.entities.sheets().get_entity('python_columns_test_{}'.format(res.ref))
        private_join_datasource_link = [link for link in sheet_with_private_join['dataSourceLinks'] if
                                        link['targetDataSourceId'] == data_source_id][0]

        outputs = [output['name'] for output in script['scriptMetadata'].get('outputs', [])]
        self.add_private_join_indicators_to_sheet(res.default_layout_id,
                                                  private_join_datasource_link['uniqueId'],
                                                  ['measure4'])

        self.commands.save_layout(res.default_layout_id)

        expected_new_df = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2, 'measure3': 3.3, 'measure4': 5.5},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22, 'measure3': 33.33, 'measure4': 55.55}]
        )

        computed_with_script_executed_df = self.kawa.sheet(force_tz='Europe/Berlin').view_id(
            res.default_layout_id).compute()

        pd_testing.assert_frame_equal(computed_with_script_executed_df.sort_index(axis=1),
                                      expected_new_df.sort_index(axis=1),
                                      check_dtype=False)

    def test_python_column_execution_with_update_of_script_to_add_inputs_and_then_execute(self):
        pk_name = 'pk strange name%%1'
        df_for_datasource = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22}]
        )
        indicator_for_param_mapping = ['measure1']
        expected_df = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2, 'measure3': 5.5},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22, 'measure3': 55.55}]
        )
        res = self.run_test_python_column_execution(df_for_datasource,
                                                    [pk_name],
                                                    'Column script measure1',
                                                    indicator_for_param_mapping,
                                                    expected_df)

        script_id = res.python_private_join['scriptId']

        # replace script by one adding one input: measure2
        script = self.commands.replace_script(script_id,
                                              'my amazing new name',
                                              'kywy-specs-final',
                                              'Columns with no pk in params')

        script_inputs = script['scriptMetadata']['parameters']

        # check input has been added to script metadata
        self.assertTrue(len(script_inputs) == 2)

        # check that error is thrown when starting python private join because of missing input
        self.assertRaisesRegex(Exception,
                               r'Cannot start the python script execution as there are some missing inputs '
                               r'defined on the Python column',
                               lambda: self.commands.start_python_private_join(res.python_private_join['id']))

        # map the new input in the PPJ
        current_mapping = res.python_private_join['paramMapping'][0]
        mapping_to_add = {
            'paramName': current_mapping['paramName'].replace('measure1', 'measure2'),
            'columnId': current_mapping['columnId'].replace('measure1', 'measure2')
        }

        self.commands.replace_python_private_join_mapping_and_prefix(
            res.python_private_join['id'],
            [current_mapping, mapping_to_add],
            'toto'
        )

        # Restart python private join
        self.commands.start_python_private_join(res.python_private_join['id'])

        # wait for the computation of the script ot finish
        time.sleep(10)

        expected_new_df = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2, 'measure3': 3.3},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22, 'measure3': 33.33}]
        )

        computed_with_script_executed_df = self.kawa.sheet(force_tz='Europe/Berlin').view_id(
            res.default_layout_id).compute()

        pd_testing.assert_frame_equal(computed_with_script_executed_df.sort_index(axis=1),
                                      expected_new_df.sort_index(axis=1),
                                      check_dtype=False)

    def test_python_column_execution_with_update_of_script_that_remove_an_input(self):
        pk_name = 'pk strange name%%1'
        df_for_datasource = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22}]
        )
        indicator_for_param_mapping = ['measure1', 'measure2']
        expected_df = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2, 'measure3': 3.3},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22, 'measure3': 33.33}]
        )
        res = self.run_test_python_column_execution(df_for_datasource,
                                                    [pk_name],
                                                    'Columns with no pk in params',
                                                    indicator_for_param_mapping,
                                                    expected_df)

        script_id = res.python_private_join['scriptId']

        # replace script by one removing one input: measure2
        script = self.commands.replace_script(script_id,
                                              'my amazing new name',
                                              'kywy-specs-final',
                                              'Column script measure1')

        script_inputs = script['scriptMetadata']['parameters']

        # check input has been added to script metadata
        self.assertTrue(len(script_inputs) == 1)

        # Start python private join
        self.commands.start_python_private_join(res.python_private_join['id'])

        # wait for the computation of the script ot finish
        time.sleep(10)

        expected_new_df = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2, 'measure3': 5.5},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22, 'measure3': 55.55}]
        )

        computed_with_script_executed_df = self.kawa.sheet(force_tz='Europe/Berlin').view_id(
            res.default_layout_id).compute()

        pd_testing.assert_frame_equal(computed_with_script_executed_df.sort_index(axis=1),
                                      expected_new_df.sort_index(axis=1),
                                      check_dtype=False)

    def test_python_column_throw_when_output_type_change(self):
        pk_name = 'pk strange name%%1'
        df_for_datasource = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22}]
        )
        indicator_for_param_mapping = ['measure1']
        expected_df = pd.DataFrame([
            {pk_name: 1, 'measure1': 1.1, 'measure2': 2.2, 'measure3': 5.5},
            {pk_name: 2, 'measure1': 11.11, 'measure2': 22.22, 'measure3': 55.55}]
        )
        res = self.run_test_python_column_execution(df_for_datasource,
                                                    [pk_name],
                                                    'Column script measure1',
                                                    indicator_for_param_mapping,
                                                    expected_df)

        script_id = res.python_private_join['scriptId']

        # replace script by one adding one input: measure2
        test_raise = lambda: self.commands.replace_script(script_id,
                                                          'my amazing new name',
                                                          'kywy-specs-final',
                                                          'Column script measure1 with ouput of different type')

        self.assertRaisesRegex(Exception,
                               r'Output with different type in repo compared to kawa',
                               test_raise)

    def test_script_with_error(self):
        df_for_datasource = pd.DataFrame([
            {'pk': 1, 'measure1': 1.1, 'measure2': 2.2},
            {'pk': 2, 'measure1': 11.11, 'measure2': 22.22}]
        )
        indicator_for_param_mapping = ['measure1', 'measure2']
        expected_df = pd.DataFrame([
            {'pk': 1, 'measure1': 1.1, 'measure2': 2.2, 'measure3': None},
            {'pk': 2, 'measure1': 11.11, 'measure2': 22.22, 'measure3': None}]
        )

        self.run_test_python_column_execution(df_for_datasource,
                                              ['pk'],
                                              'Column script with error when executing',
                                              indicator_for_param_mapping,
                                              expected_df,
                                              expected_end_status='FAILURE',
                                              expected_end_error='Test error hohoho',
                                              sleep_time_before_checking_status=60,
                                              logs='Test error hohoho')

    def test_should_throw_when_creating_script_with_two_functions_with_input_defined(self):
        self.assertRaisesRegex(Exception,
                               'More than one function decorated with @kawa_tool found in the python file associated '
                               'with this toolKit/tool. Please only decorate one function per file with @kawa_tool.',
                               self.create_script_test, 'Column script with two function will throw error')

    # def test_should_throw_when_creating_script_with_missing_arguments(self):
    #     self.assertRaisesRegex(Exception,
    #                            'Some arguments defined in the main function: execute are not defined. The list :'
    #                            ' {\'missing_param\'}.',
    #                            self.create_script_test, 'Column script with missing arguments will throw error')
    # TODO: try to do this on java side, but it's tricky

    def test_should_throw_when_creating_script_without_any_function_with_defined_inputs(self):
        self.assertRaisesRegex(Exception,
                               'No function decorated with @kawa_tool found in the python file associated with this'
                               ' toolKit/tool.',
                               self.create_script_test, 'Column script without any function with inputs')

    def add_private_join_indicators_to_sheet(self, layout_id, link_id, outputs: list):
        return self.kawa.commands._run_command(command_name='AddFieldsToLayout',
                                               command_parameters={
                                                   'layoutId': layout_id,
                                                   'fieldList': [
                                                       {
                                                           'aggregationMethod': 'FIRST',
                                                           'columnId': '{}⟶{}'.format(link_id, output),
                                                           'displayInformation': {
                                                               'displayName': output
                                                           }
                                                       }
                                                       for output in outputs
                                                   ]
                                               })

    def create_script_test(self,
                           task_name):
        df_for_datasource = pd.DataFrame([
            {'pk': 1, 'measure1': 1.1, 'measure2': 2.2},
            {'pk': 2, 'measure1': 11.11, 'measure2': 22.22}]
        )
        pks = ['pk']
        ref = uuid.uuid4()
        ds_unique_name = 'python_columns_test_{}'.format(ref)

        loader = self.kawa.new_data_loader(df=df_for_datasource, datasource_name=ds_unique_name)
        ds = loader.create_datasource(primary_keys=pks)
        loader.load_data(reset_before_insert=True, create_sheet=True)

        sheet = self.kawa.entities.sheets().get_entity(ds_unique_name)
        sheet_id = sheet['id']
        default_layout_id = sheet['defaultLayoutId']
        status = self.get_private_join_sheet_statuses(sheet_id)
        self.assertTrue(len(status) == 0)

        # share sheet and layout
        self.commands.share_sheet_and_layouts(sheet_id, [default_layout_id])

        # setup script
        script_name = 'script_{}'.format(ref)

        self.commands.create_script(script_name, 'kywy-specs-final', task_name)

    def run_test_python_column_execution(self,
                                         df_for_datasource,
                                         pks,
                                         task_name,
                                         indicator_for_param_mapping,
                                         expected_df,
                                         secrets=None,
                                         expected_end_status='SUCCESS',
                                         expected_end_error='',
                                         sleep_time_before_checking_status=10,
                                         logs=None):
        # setup datasource / sheet / grid
        ref = datetime.datetime.now().timestamp()
        if secrets is None:
            secrets = {}
        ds_unique_name = 'python_columns_test_{}'.format(ref)

        loader = self.kawa.new_data_loader(df=df_for_datasource, datasource_name=ds_unique_name)
        ds = loader.create_datasource(primary_keys=pks)
        datasource_id = ds['id']
        loader.load_data(reset_before_insert=True, create_sheet=True)

        sheet = self.kawa.entities.sheets().get_entity(ds_unique_name)
        sheet_id = sheet['id']
        default_layout_id = sheet['defaultLayoutId']
        status = self.get_private_join_sheet_statuses(sheet_id)
        self.assertTrue(len(status) == 0)

        # share sheet and layout
        self.commands.share_sheet_and_layouts(sheet_id, [default_layout_id])

        # setup script
        script_name = 'script_{}'.format(ref)

        script = self.commands.create_script(script_name,
                                             tool_kit_name='kywy-specs-final',
                                             tool_name=task_name)

        # add secrets
        if secrets:
            for k, v in secrets.items():
                self.commands.create_secret_if_not_exist(k, v, k)

        # setup private join
        script_id = script.get('id')

        python_private_join_name = 'private_join_{}'.format(ref)
        param_mapping_column_ids = ['{}⟶{}'.format(datasource_id, indicator) for indicator in
                                    indicator_for_param_mapping]
        params_from_script = script['scriptMetadata']['parameters']
        param_for_python_private_join = [{'paramName': a[0]['name'], 'columnId': a[1]} for a in
                                         zip(params_from_script, param_mapping_column_ids)]
        python_private_join = self.commands.create_python_private_join(python_private_join_name,
                                                                       sheet_id,
                                                                       default_layout_id,
                                                                       script_id,
                                                                       param_for_python_private_join)

        # add output from script to view
        join_datasource_id = python_private_join['joinDatasourceId']
        sheet_with_private_join = self.kawa.entities.sheets().get_entity(ds_unique_name)
        status = self.get_private_join_sheet_statuses(sheet_id)
        self.assertTrue(len(status) == 1)
        self.assertTrue(status[0]['pythonPrivateJoinJobStatus'] == 'MISSING_EVENT')

        script_run_histo = self.get_script_run_history(script_id)
        self.assertTrue(len(script_run_histo) == 0)

        private_join_datasource_link = [link for link in sheet_with_private_join['dataSourceLinks'] if
                                        link['targetDataSourceId'] == join_datasource_id][0]

        outputs = [output['name'] for output in script['scriptMetadata'].get('outputs', [])]
        self.add_private_join_indicators_to_sheet(default_layout_id,
                                                  private_join_datasource_link['uniqueId'],
                                                  outputs)

        self.commands.save_layout(default_layout_id)

        # when executing script
        self.commands.start_python_private_join(python_private_join['id'])

        time.sleep(sleep_time_before_checking_status)

        # then
        status = self.get_private_join_sheet_statuses(sheet_id)
        self.assertTrue(len(status) == 1)
        self.assertTrue(status[0]['pythonPrivateJoinJobStatus'] == expected_end_status)
        self.assertTrue(expected_end_error in status[0].get('errorMessage', ''))

        script_run_histo_end = self.get_script_run_history(script_id)
        self.assertTrue(len(script_run_histo_end) == 1)
        self.assertTrue(script_run_histo_end[0]['status'] == expected_end_status)
        self.assertTrue(expected_end_error in script_run_histo_end[0].get('errorMessage', ''))

        # if not FAILURE check the script data
        if expected_end_status != 'FAILURE':
            computed_with_script_executed_df = self.kawa.sheet(force_tz='Europe/Berlin').view_id(
                default_layout_id).compute()
            pd_testing.assert_frame_equal(computed_with_script_executed_df.sort_index(axis=1),
                                          expected_df.sort_index(axis=1),
                                          check_dtype=False)

        if logs:
            job_execution_id = script_run_histo_end[0].get('jobExecutionId')
            logs_from_server = self.get_job_log(job_execution_id['relatedEntityId'], job_execution_id['contextId'])
            self.assertTrue(logs in logs_from_server)

        return TestResult(python_private_join, ref, datasource_id, default_layout_id, sheet_id)

    def get_private_join_sheet_statuses(self, sheet_id: str) -> list:
        return self.kawa.get(
            '{}/backoffice/sheets/{}/python-private-joins-current-status'.format(self.kawa.kawa_api_url, sheet_id))

    def get_script_run_history(self, script_id: str) -> list:
        return self.kawa.get(
            '{}/backoffice/scripts/{}/run-history'.format(self.kawa.kawa_api_url, script_id))

    def get_job_log(self, entity_id: str, execution_id: str) -> str:
        return self.kawa.get(
            '{}/backoffice/joblog/{}/{}'.format(self.kawa.kawa_api_url, entity_id, execution_id)).get('log', '')


@dataclass
class TestResult:
    """Class for keeping track of an item in inventory."""
    python_private_join: dict
    ref: uuid.UUID
    datasource_id: str
    default_layout_id: str
    sheet_id: str


def all_type_row(pk: int, f: float, i: int, d: datetime.date, dt: datetime.datetime, b: bool, s: str) -> dict:
    return {'pk': pk, 'decimal': f, 'integer': i, 'date': d, 'date_time': pytz.timezone('Europe/Berlin').localize(dt),
            'boolean': b, 'text': s}


def all_type_row_res(f: float, i: int, d: datetime.date, dt: datetime.datetime, b: bool, s: str) -> dict:
    return {'decimal_res': f, 'integer_res': i, 'date_res': d,
            'datetime_res': pytz.timezone('Europe/Berlin').localize(dt), 'boolean_res': b, 'text_res': s}


def at_berlin(dt: datetime.datetime):
    return pytz.timezone('Europe/Berlin').localize(dt)



import os.path
from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED
import uuid
import glob
import shutil
import numpy as np
import time


class KawaData:

    def __init__(self, kawa_client):
        self._k = kawa_client

    def wrap_sql_query(self, sheet_name_or_id, sql):
        sheet = self._k.entities.sheets().get_entity(sheet_name_or_id)
        if not sheet:
            raise Exception('Sheet with name/id {} was not found in the current workspace'.format(sheet_name_or_id))

        url = '{}/computation/sql-query-for-gen-ai'.format(self._k.kawa_api_url)
        response = self._k.post(url=url, data={
            'sheetName': sheet.get('displayInformation').get('displayName'),
            'sql': sql
        })
        return response.get('sql')

    def create_datasource_and_load(self,
                                   datasource_name,
                                   df,
                                   column_to_attribute=None,
                                   primary_key_columns=None,
                                   ignore_new_columns=False,
                                   default_columns=None,
                                   create_sheet=False,
                                   reset_data=True,
                                   delete_datasource_and_recreate=False,
                                   type_overrides=None,
                                   shared=False,
                                   list_of_parquet_files=None,
                                   **xargs):

        return self.__create_datasource_and_load(datasource_name=datasource_name,
                                                 df=df,
                                                 column_to_attribute=column_to_attribute,
                                                 primary_key_columns=primary_key_columns,
                                                 ignore_new_columns=ignore_new_columns,
                                                 default_columns=default_columns,
                                                 create_sheet=create_sheet,
                                                 reset_data=reset_data,
                                                 delete_datasource_and_recreate=delete_datasource_and_recreate,
                                                 type_overrides=type_overrides,
                                                 shared=shared,
                                                 list_of_parquet_files=list_of_parquet_files,
                                                 private_join_datasource=False)

    def create_private_join_datasource_and_load(self,
                                                datasource_name,
                                                df,
                                                column_to_attribute=None,
                                                default_columns=None,
                                                type_overrides=None,
                                                show_progress=True,
                                                list_of_parquet_files=None):

        self.__create_datasource_and_load(datasource_name=datasource_name,
                                          df=df,
                                          column_to_attribute=column_to_attribute,
                                          primary_key_columns=None,
                                          ignore_new_columns=False,
                                          default_columns=default_columns,
                                          create_sheet=False,
                                          reset_data=True,
                                          delete_datasource_and_recreate=False,
                                          type_overrides=type_overrides,
                                          shared=False,
                                          show_progress=show_progress,
                                          list_of_parquet_files=list_of_parquet_files,
                                          private_join_datasource=True)

    def __create_datasource_and_load(self,
                                     datasource_name,
                                     df,
                                     column_to_attribute=None,
                                     primary_key_columns=None,
                                     ignore_new_columns=False,
                                     default_columns=None,
                                     create_sheet=False,
                                     reset_data=True,
                                     delete_datasource_and_recreate=False,
                                     type_overrides=None,
                                     shared=False,
                                     show_progress=True,
                                     list_of_parquet_files=None,
                                     private_join_datasource=False):

        datasource_id = self._k.entities.datasources().get_entity_id(entity_name=datasource_name)

        if not primary_key_columns:
            primary_key_columns = []
        if not column_to_attribute:
            column_to_attribute = {}

        if datasource_id and delete_datasource_and_recreate:
            self._k.commands.delete_datasource(datasource=datasource_id)

        datasource = self.__create_datasource(df=df,
                                              shared=shared,
                                              default_columns=default_columns,
                                              datasource_name=datasource_name,
                                              column_to_attribute=column_to_attribute,
                                              primary_key_columns=primary_key_columns,
                                              type_overrides=type_overrides,
                                              private_join_datasource=private_join_datasource)

        if not ignore_new_columns:
            self.__add_missing_indicators_to_datasource(df=df,
                                                        datasource=datasource,
                                                        type_overrides=type_overrides)

        self.__load_data(df=df,
                         datasource=datasource,
                         reset_data=reset_data,
                         show_progress=show_progress,
                         list_of_parquet_files=list_of_parquet_files)

        if create_sheet:
            self._k.commands.create_sheet(datasource=datasource,
                                          sheet_name=datasource_name)

        return datasource

    def __load_data(self, datasource, df, show_progress=True, reset_data=True, list_of_parquet_files=None):
        print(
            'WARNING: This method to load data is deprecated. '
            'Please refer to https://docs.kawa.ai/python-api-for-data-analytics#36e18270b6c9465b95252248ff3802b5')
        datasource_id = datasource if isinstance(datasource, str) else datasource.get('id')
        indicators = datasource.get('indicators')
        session_id = str(uuid.uuid4())

        print('Starting an ingestion session with id={}'.format(session_id))

        # URLs for ingestion session
        query_params = 'datasource={}&format=parquet&reset={}&session={}'.format(datasource_id, reset_data, session_id)
        prepare_url = '{}/ingestion/prepare?{}'.format(self._k.kawa_api_url, query_params)
        ingest_url = '{}/ingestion/upload?{}'.format(self._k.kawa_api_url, query_params)
        finalize_url = '{}/ingestion/finalize?{}'.format(self._k.kawa_api_url, query_params)
        finalize_for_failure_url = '{}/ingestion/stop-with-failure?{}'.format(self._k.kawa_api_url, query_params)

        # Check that all date and date time indicators are numbers in the data frame
        temporal_indicators = [i for i in indicators if
                               i.get('type') == 'date' or i.get('type') == 'date_time']
        for temporal_indicator in temporal_indicators:
            df_type_name = df[temporal_indicator.get('indicatorId')].dtype
            if 'datetime' in str(df_type_name):
                raise Exception(
                    "When using accelerated loading, "
                    "date and date_time columns should be converted to numbers. Date columns should be expressed "
                    "as number of days since epoch and Date time columns as number of milliseconds since epoch.")

        # Call prepare data that will check if we can start loading and give us the offset for automatic index
        prepare_data = self._k.post(url=prepare_url, data={})

        if not prepare_data.get('canRunLoading'):
            raise Exception(
                'We cannot start ingestion due to: ' + prepare_data.get('raisonItCannotStart', 'No reason given'))

        parquet_directory = '{}/{}'.format(self._k.tmp_files_directory, str(uuid.uuid4()))
        os.makedirs(parquet_directory, exist_ok=True)

        try:
            auto_increment_indicator = [i for i in indicators if
                                        i.get('storageConfig', {}).get('automaticUniqueValue', False)]

            if len(auto_increment_indicator) == 1:
                if 'offsetToApplyToAutoIncrementIndex' not in prepare_data:
                    self._k.post(url=finalize_for_failure_url, data={})
                    raise Exception('The offset for to the auto_increment_index was not present in the answer from '
                                    'backend. Cannot continue')

                auto_increment_indicator_id = auto_increment_indicator[0].get('indicatorId')
                df[auto_increment_indicator_id] = df.index + prepare_data.get('offsetToApplyToAutoIncrementIndex') + 1

            if list_of_parquet_files:
                futures = []
                with ThreadPoolExecutor(10) as executor:
                    for file_name in list_of_parquet_files:
                        futures.append(executor.submit(self._k.post_binary_file, file_name, ingest_url))
                wait(futures, return_when=ALL_COMPLETED)

            else:
                partition_cols = []

                if show_progress:
                    print('> Exporting the dataframe a parquet file')
                df.to_parquet(partition_cols=partition_cols, path=parquet_directory + '/', compression='gzip')
                start = time.time()
                file_name = glob.glob('{}/**/*.parquet'.format(parquet_directory), recursive=True)[0]
                self._k.post_binary_file(filename=file_name, url=ingest_url)
                end = time.time()
                if show_progress:
                    print('> {} rows were imported in {}ms'.format(df.shape[0], end - start))

        except Exception as e:
            self._k.post(url=finalize_for_failure_url, data={})
            raise e

        finally:
            if os.path.isdir(parquet_directory):
                shutil.rmtree(parquet_directory)

            if list_of_parquet_files:
                for filename in list_of_parquet_files:
                    try:
                        os.remove(filename)
                    except OSError:
                        pass

            self._k.post(url=finalize_url, data={})
            if show_progress:
                print('> Import was successfully finalized')

    def __add_missing_indicators_to_datasource(self, datasource, df, type_overrides=None):

        if not type_overrides:
            type_overrides = {}

        existing_indicator_id = [i.get('indicatorId') for i in datasource.get('indicators', [])]

        new_indicators = []
        for column_name in df.columns:
            new_indicator = self.__get_indicator_from_column(type_override=type_overrides.get(column_name),
                                                             df_column=df[column_name])

            new_indicator_id = new_indicator.get('indicatorId')
            if not (new_indicator_id in existing_indicator_id):
                new_indicators.append(new_indicator)

        if new_indicators:
            print('Adding the following indicators: {}'.format([i.get('indicatorId') for i in new_indicators]))
            self._k.commands.add_indicators_to_datasource(datasource=datasource,
                                                          new_indicators=new_indicators)

    def __create_datasource(self, datasource_name, df, column_to_attribute, primary_key_columns,
                            shared=False, default_columns=None, type_overrides=None, private_join_datasource=False):

        if not type_overrides:
            type_overrides = {}

        # Create all the needed attributes first
        # (This operation is idempotent)
        for key, value in column_to_attribute.items():
            self._k.commands.create_attribute(value)

        indicators = []
        for column_name in df.columns:
            is_attribute = column_name in column_to_attribute
            is_default = not default_columns or column_name in default_columns

            indicator = self.__get_indicator_from_column(df_column=df[column_name],
                                                         type_override=type_overrides.get(column_name),
                                                         attribute_name=column_to_attribute.get(column_name),
                                                         is_default=is_default or is_attribute,
                                                         is_primary_key=column_name in primary_key_columns)
            indicators.append(indicator)

        # A private join datasource must have the same primary keys as the main datasource
        if private_join_datasource:
            main_datasource = self._k.entities.datasources().get_entity(entity_name=datasource_name)
            for main_ds_indicator in main_datasource['indicators']:
                if 'key' in main_ds_indicator:
                    key = main_ds_indicator['key']
                    if 'keyType' in key:
                        keyType = key['keyType']
                        if keyType == 'PRIMARY_SHARDING_KEY' or keyType == 'PRIMARY_KEY':
                            indicators.append(main_ds_indicator)

        # Add the auto increment key if there is no specified key
        key_indicators = [i for i in indicators if 'key' in i]
        if not key_indicators and not private_join_datasource:
            indicators.append({
                'displayInformation': {
                    'displayName': 'record_id'
                },
                'includedInDefaultLayout': False,
                'indicatorId': 'record_id',
                'storageConfig': {
                    'indexed': True,
                    'automaticUniqueValue': True
                },
                'type': 'integer',
                'key': {
                    'keyType': 'PRIMARY_SHARDING_KEY'
                }
            })

        datasource = {
            'shared': shared,
            'displayInformation': {
                'displayName': datasource_name
            },
            'storageConfiguration': {
                'loadingAdapterName': 'CLICKHOUSE'
            },
            'indicators': indicators,
            'dataCategoryId': '1'
        }

        if private_join_datasource:
            return self._k.commands.create_private_join_datasource(datasource)
        else:
            return self._k.commands.create_datasource(datasource)

    def __get_indicator_from_column(self, df_column, is_default=True, type_override=None, attribute_name=None,
                                    is_primary_key=False):

        attribute_id = (self._k
                        .entities
                        .attributes()
                        .get_entity_id(attribute_name))

        column_name = df_column.name
        df_type_name = df_column.dtype.base
        kawa_type = self.__resolve_type(df_type_name.name)

        # We have to detect arrays as well as their types
        # This is a poor method but let's leave it there for now
        first_defined_element = None
        for element in df_column.to_list():
            if element is not None:
                first_defined_element = element
                break

        if isinstance(first_defined_element, list):
            list_from_df = first_defined_element
            if len(list_from_df) == 0:
                kawa_type = 'list(integer,text)'
            else:
                first = list_from_df[0]
                if isinstance(first, (int, float, complex)) and not isinstance(first, bool):
                    kawa_type = 'list(integer,decimal)'
                else:
                    kawa_type = 'list(integer,text)'

        # Apply override if available
        if type_override:
            kawa_type = type_override

        indicator = {
            'displayInformation': {
                'displayName': column_name
            },
            'includedInDefaultLayout': is_default,
            'indicatorId': column_name,
            'storageConfig': {
                'indexed': bool(is_primary_key or attribute_id)
            },
            'type': kawa_type
        }

        if attribute_id is not None:
            indicator['attributeId'] = attribute_id

        if is_primary_key:
            indicator['key'] = {
                'keyType': 'PRIMARY_SHARDING_KEY'
            }

        return indicator

    @staticmethod
    def __resolve_type(np_type_name):
        if np_type_name in KAWA_TYPES_DICT:
            return KAWA_TYPES_DICT.get(np_type_name)

        if 'datetime' in np_type_name:
            return K_DATE

        return K_TEXT


K_CATEGORY = 'category'
K_DATE = 'date'
K_BOOLEAN = 'boolean'
K_TEXT = 'text'
K_DECIMAL = 'decimal'
K_INTEGER = 'integer'

KAWA_TYPES_DICT = {np.dtype(str).name: K_TEXT,
                   np.dtype(object).name: K_TEXT,
                   K_CATEGORY: K_TEXT,
                   np.dtype(bool).name: K_BOOLEAN,
                   np.dtype(np.datetime64).name: K_DATE,
                   np.dtype(np.int8).name: K_INTEGER,
                   np.dtype(np.int16).name: K_INTEGER,
                   np.dtype(np.int32).name: K_INTEGER,
                   np.dtype(np.int64).name: K_INTEGER,
                   np.dtype(np.timedelta64).name: K_INTEGER,
                   np.dtype(np.uint8).name: K_INTEGER,
                   np.dtype(np.uint16).name: K_INTEGER,
                   np.dtype(np.uint32).name: K_INTEGER,
                   np.dtype(np.uint64).name: K_INTEGER,
                   np.dtype(np.float16).name: K_DECIMAL,
                   np.dtype(np.float32).name: K_DECIMAL,
                   np.dtype(np.float64).name: K_DECIMAL,
                   np.dtype(np.complex64).name: K_DECIMAL,
                   np.dtype(np.complex128).name: K_DECIMAL
                   }



from ...kawa_base_e2e_test import KawaBaseTest
import pandas as pd
import uuid
import tempfile
from ....client.kawa_client import KawaClient as K


class TestComputationWithSorting(KawaBaseTest):
    cities_and_countries = pd.DataFrame([
        {'id': 'a', 'country': 'FR', 'city': 'Paris', 'measure': 1},
        {'id': 'b', 'country': 'FR', 'city': 'Lyon', 'measure': 2},
        {'id': 'c', 'country': 'FR', 'city': 'Lille', 'measure': 3},
        {'id': 'd', 'country': 'UK', 'city': 'London', 'measure': 4},
        {'id': 'e', 'country': 'UK', 'city': 'Cardiff', 'measure': 5},
        {'id': 'g', 'country': 'UK', 'city': 'Edinburgh', 'measure': 6},
        {'id': 'h', 'country': 'UK', 'city': 'Belfast', 'measure': 7},
        {'id': 'i', 'country': 'BE', 'city': 'Brussels', 'measure': 8},
        {'id': 'j', 'country': 'BE', 'city': 'Bruges', 'measure': 9},
        {'id': 'k', 'country': 'BE', 'city': 'Namur', 'measure': 10},
    ])

    cities = ['Paris', 'Lyon', 'Lille', 'London', 'Cardiff', 'Edinburgh', 'Belfast', 'Brussels', 'Bruges', 'Namur']

    @classmethod
    def setUpClass(cls):
        unique_id = 'resource_{}'.format(uuid.uuid4())
        print(tempfile.gettempdir())
        KawaBaseTest.setUpClass()
        cls.sheet_name = unique_id

    def test_that_sort_bottom_1_can_be_extracted_when_there_is_no_grouping(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.cities_and_countries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa.sheet(sheet_name=unique_id)
              .order_by('city', ascending=True)
              .select(K.col('city'))
              .limit(1)
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['city'])
        self.assertListEqual(list(df['city']), ['Belfast'])

    def test_that_sort_bottom_group_can_be_extracted(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.cities_and_countries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa.sheet(sheet_name=unique_id)
              .order_by('measure', ascending=True)
              .select(K.col('measure').sum())
              .group_by('country')
              .limit(1)
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['grouping(0)', 'measure'])
        self.assertListEqual(list(df['grouping(0)']), ['FR'])
        self.assertListEqual(list(df['measure']), [6])

    def test_that_sort_top_1_can_be_extracted_when_there_is_no_grouping(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.cities_and_countries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa.sheet(sheet_name=unique_id)
              .order_by('city', ascending=False)
              .select(K.col('city'))
              .limit(1)
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['city'])
        self.assertListEqual(list(df['city']), ['Paris'])

    def test_that_sort_top_group_can_be_extracted(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.cities_and_countries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa.sheet(sheet_name=unique_id)
              .order_by('measure', ascending=False)
              .select(K.col('measure').sum())
              .group_by('country')
              .limit(1)
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['grouping(0)', 'measure'])
        self.assertListEqual(list(df['grouping(0)']), ['BE'])
        self.assertListEqual(list(df['measure']), [27])

    def test_that_sort_ascending_works_without_group_by(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.cities_and_countries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa.sheet(sheet_name=unique_id)
              .order_by('city', ascending=True)
              .select(K.col('city'))
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['city'])
        self.cities.sort()
        self.assertListEqual(list(df['city']), self.cities)

    def test_that_sort_descending_works_without_group_by(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.cities_and_countries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa.sheet(sheet_name=unique_id)
              .order_by('city', ascending=False)
              .select(K.col('city'))
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['city'])
        self.cities.sort(reverse=True)
        self.assertListEqual(list(df['city']), self.cities)

    def test_that_sort_ascending_works_with_group_by(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.cities_and_countries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa.sheet(sheet_name=unique_id)
              .group_by('country')
              .order_by('measure', ascending=True)
              .select(K.col('measure').sum())
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['grouping(0)', 'measure'])
        self.assertListEqual(list(df['measure']), [6, 22, 27])
        self.assertListEqual(list(df['grouping(0)']), ['FR', 'UK', 'BE'])

    def test_that_sort_descending_works_with_group_by(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.cities_and_countries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa.sheet(sheet_name=unique_id)
              .group_by('country')
              .order_by('measure', ascending=False)
              .select(K.col('measure').sum())
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['grouping(0)', 'measure'])
        self.assertListEqual(list(df['measure']), [27, 22, 6])
        self.assertListEqual(list(df['grouping(0)']), ['BE', 'UK', 'FR'])



import pandas as pd
import uuid
from datetime import datetime, time
from ...test.kawa_base_e2e_test import KawaBaseTest
from ...client.kawa_client import KawaClient as kawa
import zoneinfo


class TestComputationDslWithVariousDataSets(KawaBaseTest):
    eastern_tz = zoneinfo.ZoneInfo('US/Eastern')
    utc_tz = zoneinfo.ZoneInfo('UTC')

    cities_and_countries = pd.DataFrame([
        {'id': 'a', 'country': 'FR', 'city': 'Paris', 'measure': 1},
        {'id': 'b', 'country': 'FR', 'city': 'Lyon', 'measure': 2},
        {'id': 'c', 'country': 'FR', 'city': 'Lille', 'measure': 3},
        {'id': 'd', 'country': 'UK', 'city': 'London', 'measure': 4},
        {'id': 'e', 'country': 'UK', 'city': 'Cardiff', 'measure': 5},
        {'id': 'g', 'country': 'UK', 'city': 'Edinburgh', 'measure': 6},
        {'id': 'h', 'country': 'UK', 'city': 'Belfast', 'measure': 7},
        {'id': 'i', 'country': 'BE', 'city': 'Brussels', 'measure': 8},
        {'id': 'j', 'country': 'BE', 'city': 'Bruges', 'measure': 9},
        {'id': 'k', 'country': 'BE', 'city': 'Namur', 'measure': 10},
    ])
    market_data = pd.DataFrame([
        {'id': 1, 'ticker': 'AAPL', 'price': 13.3, 'time': datetime(2024, 1, 1, 13, 14, 1, tzinfo=eastern_tz)},
        {'id': 2, 'ticker': 'AAPL', 'price': 12.4, 'time': datetime(2024, 1, 1, 13, 14, 34, tzinfo=eastern_tz)},
        {'id': 3, 'ticker': 'AAPL', 'price': 12.5, 'time': datetime(2024, 1, 1, 15, 14, 5, tzinfo=eastern_tz)},
        {'id': 4, 'ticker': 'AAPL', 'price': 12.7, 'time': datetime(2024, 1, 1, 15, 15, 43, tzinfo=eastern_tz)},
        {'id': 5, 'ticker': 'MSFT', 'price': 23.3, 'time': datetime(2024, 1, 1, 13, 14, 1, tzinfo=eastern_tz)},
        {'id': 6, 'ticker': 'MSFT', 'price': 22.4, 'time': datetime(2024, 1, 1, 13, 14, 34, tzinfo=eastern_tz)},
        {'id': 7, 'ticker': 'MSFT', 'price': 22.5, 'time': datetime(2024, 1, 1, 15, 14, 5, tzinfo=eastern_tz)},
        {'id': 8, 'ticker': 'MSFT', 'price': 22.7, 'time': datetime(2024, 1, 1, 15, 15, 43, tzinfo=eastern_tz)},
    ])
    tiny_timeseries = pd.DataFrame([
        {'id': 13, 'time': datetime(2021, 1, 4, 13, 0, 0, tzinfo=eastern_tz)},
        {'id': 14, 'time': datetime(2022, 1, 3, 14, 0, 0, tzinfo=eastern_tz)},
        {'id': 15, 'time': datetime(2023, 1, 4, 15, 0, 0, tzinfo=eastern_tz)},
        {'id': 16, 'time': datetime(2024, 1, 1, 16, 0, 0, tzinfo=eastern_tz)},
    ])

    def test_group_by_1(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.market_data, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa.sheet(sheet_name=unique_id)
              .group_by('1')
              .select(kawa.col('price').sum(), kawa.col('id').count())
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['grouping(0)', 'price', 'id'])
        self.assertListEqual(list(df['grouping(0)']), ['1'])
        self.assertListEqual(list(df['price']), [13.3 + 12.4 + 12.5 + 12.7 + 23.3 + 22.4 + 22.5 + 22.7])
        self.assertListEqual(list(df['id']), [8])

    def test_time_range_filters(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.tiny_timeseries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        def compute(tz_id,
                    from_inclusive=None,
                    from_exclusive=None,
                    to_inclusive=None,
                    to_exclusive=None):
            time_filter = kawa.col('time').time_range(from_inclusive=from_inclusive,
                                                      from_exclusive=from_exclusive,
                                                      to_inclusive=to_inclusive,
                                                      to_exclusive=to_exclusive)
            return list((self.kawa
                         .sheet(sheet_name=unique_id, force_tz=tz_id)
                         .select('id')
                         .filter(time_filter)
                         .compute())['id'])

        # then
        # Data contains
        # 13:00, 14:00, 15:00, 16:00 in eastern time
        # Which translates to
        # 18:00, 19:00, 20:00, 21:00 in UTC

        # ----------------
        # US/Eastern computations
        # ----------------
        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_inclusive=time(hour=15, minute=0, second=0)
            ), [15, 16])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_exclusive=time(hour=15, minute=0, second=0)
            ), [16])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                to_inclusive=time(hour=15, minute=0, second=0)
            ), [13, 14, 15])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                to_exclusive=time(hour=15, minute=0, second=0)
            ), [13, 14])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_exclusive=time(hour=13, minute=0, second=0),
                to_exclusive=time(hour=15, minute=0, second=0),
            ), [14])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_inclusive=time(hour=13, minute=0, second=0),
                to_exclusive=time(hour=15, minute=0, second=0),
            ), [13, 14])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_exclusive=time(hour=13, minute=0, second=0),
                to_inclusive=time(hour=15, minute=0, second=0),
            ), [14, 15])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_inclusive=time(hour=15, minute=0, second=0),
                to_inclusive=time(hour=15, minute=0, second=0),
            ), [15])

        self.assertListEqual(
            compute(
                tz_id='US/Eastern',
                from_exclusive=time(hour=15, minute=0, second=0),
                to_inclusive=time(hour=15, minute=0, second=0),
            ), [])

        # ----------------
        # UTC computations
        # ----------------
        self.assertListEqual(
            compute(
                tz_id='UTC',
                # Equivalent of 15:00 Eastern
                from_inclusive=time(hour=20, minute=0, second=0)
            ), [15, 16])

        self.assertListEqual(
            compute(
                tz_id='UTC',
                # Equivalent of 15:00 Eastern
                from_exclusive=time(hour=20, minute=0, second=0)
            ), [16])

        self.assertListEqual(
            compute(
                tz_id='UTC',
                # Equivalent of 15:00 Eastern
                to_inclusive=time(hour=20, minute=0, second=0)
            ), [13, 14, 15])

        self.assertListEqual(
            compute(
                tz_id='UTC',
                # Equivalent of 15:00 Eastern
                to_exclusive=time(hour=20, minute=0, second=0)
            ), [13, 14])

    def test_that_computation_works_with_two_levels_of_grouping_and_datetime_up_sampling(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.market_data, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(
            reset_before_insert=True,
            create_sheet=True)

        # when
        df = (self.kawa.sheet(sheet_name=unique_id)
              .sample(column_name='time', sampler='FIFTEEN_MINUTES')
              .group_by('time', 'ticker')
              .agg(kawa.col('price').sum())
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['grouping(0)', 'grouping(1)', 'price'])
        self.assertListEqual(list(df['price']), [
            13.3 + 12.4, 23.3 + 22.4,
            12.5, 22.5,
            12.7, 22.7,
        ])
        actual_ts = [ts.timestamp() for ts in df['grouping(0)']]
        expected_ts = [
            datetime(2024, 1, 1, 13, 0, 0, tzinfo=self.eastern_tz).timestamp(),
            datetime(2024, 1, 1, 13, 0, 0, tzinfo=self.eastern_tz).timestamp(),
            datetime(2024, 1, 1, 15, 0, 0, tzinfo=self.eastern_tz).timestamp(),
            datetime(2024, 1, 1, 15, 0, 0, tzinfo=self.eastern_tz).timestamp(),
            datetime(2024, 1, 1, 15, 15, 0, tzinfo=self.eastern_tz).timestamp(),
            datetime(2024, 1, 1, 15, 15, 0, tzinfo=self.eastern_tz).timestamp(),
        ]
        self.assertListEqual(actual_ts, expected_ts)
        self.assertListEqual(list(df['grouping(1)']), ['AAPL', 'MSFT', 'AAPL', 'MSFT', 'AAPL', 'MSFT'])

    def test_that_computation_works_with_two_levels_of_grouping(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.cities_and_countries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa.sheet(sheet_name=unique_id)
              .group_by('country', 'city')
              .agg(kawa.col('measure').sum())
              .compute())

        # then
        self.assertListEqual(list(df.columns), ['grouping(0)', 'grouping(1)', 'measure'])
        self.assertListEqual(list(df['grouping(0)']), [
            'BE', 'BE', 'BE',
            'FR', 'FR', 'FR',
            'UK', 'UK', 'UK', 'UK'])
        self.assertListEqual(list(df['grouping(1)']), [
            'Bruges', 'Brussels', 'Namur',
            'Lille', 'Lyon', 'Paris',
            'Belfast', 'Cardiff', 'Edinburgh', 'London'
        ])

    def test_that_computation_works_with_two_levels_of_grouping_keeping_group_names(self):
        # setup
        unique_id = 'resource_{}'.format(uuid.uuid4())
        loader = self.kawa.new_data_loader(df=self.cities_and_countries, datasource_name=unique_id)
        loader.create_datasource(primary_keys=['id'])
        loader.load_data(reset_before_insert=True, create_sheet=True)

        # when
        df = (self.kawa.sheet(sheet_name=unique_id)
              .group_by('country', 'city')
              .agg(kawa.col('measure').sum())
              .compute(use_group_names=True))

        # then
        self.assertListEqual(list(df.columns), ['group(0) country', 'group(1) city', 'measure'])
        self.assertListEqual(list(df['group(0) country']), [
            'BE', 'BE', 'BE',
            'FR', 'FR', 'FR',
            'UK', 'UK', 'UK', 'UK'])
        self.assertListEqual(list(df['group(1) city']), [
            'Bruges', 'Brussels', 'Namur',
            'Lille', 'Lyon', 'Paris',
            'Belfast', 'Cardiff', 'Edinburgh', 'London'
        ])



class KawaEntities:

    def __init__(self, kawa_client):
        self._k = kawa_client
        self._entity_kind = None

    def entities_of_kind(self, entity_kind):
        self._entity_kind = entity_kind
        return self

    def conversations(self):
        self._entity_kind = 'conversations'
        return self

    def perimeters(self):
        self._entity_kind = 'perimeters'
        return self

    def attributes(self):
        self._entity_kind = 'attributes'
        return self

    def layouts(self):
        self._entity_kind = 'layouts'
        return self

    def data_categories(self):
        self._entity_kind = 'data-categories'
        return self

    def data_providers(self):
        self._entity_kind = 'data-providers'
        return self

    def datasources(self):
        self._entity_kind = 'datasources'
        return self

    def dashboards(self):
        self._entity_kind = 'dashboards'
        return self

    def sheets(self):
        self._entity_kind = 'sheets'
        return self

    def views(self):
        self._entity_kind = 'extended-views'
        return self

    def principals(self):
        self._entity_kind = 'principals'
        return self

    def workspaces(self):
        self._entity_kind = 'workspaces'
        return self

    def script_runners(self):
        self._entity_kind = 'script-runners'
        return self

    def secrets(self):
        self._entity_kind = 'secret'
        return self

    def python_private_joins(self):
        self._entity_kind = 'python-private-joins'
        return self

    def agents(self):
        self._entity_kind = 'agents'
        return self

    def list_entities(self):
        return self._k.get('{}/backoffice/{}'.format(self._k.kawa_api_url, self._entity_kind.lower()))

    def get_entity_by_id(self, id):
        return self._k.get('{}/backoffice/{}/{}'.format(self._k.kawa_api_url, self._entity_kind.lower(), id))

    def list_entity_names(self):
        return [e.get('displayInformation').get('displayName') for e in self.list_entities()]

    def get_datasource_schema(self, datasource_id=None):
        return self._k.get('{}/backoffice/datasources/{}/schema'.format(self._k.kawa_api_url, datasource_id))

    def get_sheet_schema(self, sheet_id=None):
        return (self._k
                .get('{}/backoffice/sheets/{}/schema'.format(self._k.kawa_api_url, sheet_id))
                .get('schema'))

    def get_sheet_parameters(self, sheet_id=None):
        control_panel = self._k.get(f'{self._k.kawa_api_url}/backoffice/extended-control-panels/sheets/{sheet_id}')

        all_parameters = []
        for control in control_panel.get('controls', []):
            parameter_id = control.get('parameterId')
            control_name = control.get('displayInformation').get('displayName')

            for parameter in control_panel.get('parameters', []):
                if parameter.get('id') == parameter_id:
                    parameter_type = parameter.get('type')
                    all_parameters.append({
                        'type': parameter_type,
                        'name': control_name,
                        'id': parameter_id
                    })

        return all_parameters

    def get_entity(self, entity_id_or_name=None):
        if not entity_id_or_name:
            return None

        entity_id_or_name = str(entity_id_or_name)

        if self._entity_kind == 'attributes':
            # Attributes are saved in upper case so if we want to search for some,
            # we need to translate the searched for name to upper case as well.
            entity_name = entity_id_or_name.upper()

        entities = self.list_entities()
        for entity in entities:
            candidate_name = entity.get('displayInformation').get('displayName')
            candidate_id = entity.get('id')
            if candidate_name == entity_id_or_name or candidate_id == entity_id_or_name:
                return entity

        return None

    def get_entity_id(self, entity_name):
        entity = self.get_entity(entity_name)
        if entity:
            return entity.get('id')

    def find_first_entity_by_attribute(self, attribute, value):
        for entity in self.list_entities():
            if entity.get(attribute) == value:
                return entity

    def find_entities_by_attribute(self, attribute, value_list):
        res = []
        value_set = set(value_list)
        for entity in self.list_entities():
            if entity.get(attribute) in value_set:
                res.append(entity)
        return res



from random import randint
from datetime import datetime

class Generators:

    @staticmethod
    def get_random_element_in(elements):
        rand_index = randint(0, len(elements) - 1)
        return elements[rand_index]

    @staticmethod
    def generate_random_int(lower_bound, upper_bound):
        return randint(lower_bound, upper_bound)

    @staticmethod
    def generate_weighted_list(elements_with_weight):
        weighed_list = []
        for element in elements_with_weight:
            weighed_list.extend([element[0]] * element[1])
        return weighed_list

    @staticmethod
    def generate_number_of_days_since_epoch(year, month, day):
        d = datetime(year, month, day, 0, 0, 0)
        return (d - datetime(1970, 1, 1)).days

    @staticmethod
    def generate_timestamp(year, month, day, hour, minute, second):
        d = datetime(year, month, day, hour, minute, second)
        return int(d.timestamp() * 1000)

    @staticmethod
    def generate_random_names(limit=1000):
        return [name.strip() for name in """
        Amanda
        Burns Sharon
        Fuentes Michael
        Watson Cynthia
        Martin Amber
        Hopkins Robert
        Edwards Judy
        Moore Kelly
        Carter Anthony
        Smith Margaret
        Martin Keith
        Wright Stephanie
        Anderson Ricky
        Alvarado John
        Galvan Michael
        Johnson Dana
        Burch Sonya
        Carroll David
        Kline Christopher
        Grant Brittany
        Anderson Travis
        Wagner Kelly
        Rojas Lance
        Payne Susan
        Morgan Timothy
        Harrison Taylor
        Bowen Daniel
        Powers Michael
        Garrison Jennifer
        Byrd Katherine
        Williams Debra
        Cook Brendan
        Lopez Miranda
        Moore Larry
        Griffith Louis
        Flowers James
        Williams Jacob
        Cordova James
        Perry James
        Gonzalez Catherine
        Roberts Lisa
        Andrews William
        Cunningham
        DVM Corey
        Vargas Jimmy
        Werner Natalie
        Franklin Sheila
        Romero Catherine
        Garcia Jodi
        Trevino Erin
        Hart Lisa
        Lowe Joseph
        Morrison Michael
        Quinn Danielle
        Mejia Brian
        Matthews Mr.Eric
        Freeman Karen
        Robertson Kelly
        Montoya Tina
        Murray Charles
        Nunez Larry
        Kramer Margaret
        Christian Daniel
        Wood John
        Edwards Steven
        Guerrero Martin
        Williams John
        Rogers Jimmy
        Hunter Sharon
        Smith Michele
        Deleon Kelly
        Blair Daniel
        Mckinney Nathaniel
        Lowery Kevin
        Hunt Mary
        Terry Sandra
        Cisneros Eric
        Ortega Connie
        James Jennifer
        Rodriguez Joann
        Reed Lisa
        Pope Thomas
        Montgomery
        MD Brad
        Conway Philip
        Reynolds Brian
        Howard Allen
        Turner Misty
        Miller Nancy
        Spencer Brandy
        Holloway Michael
        Berry Cynthia
        Alvarez Brandon
        Ward Keith
        Morrison Benjamin
        Martin Kevin
        Lloyd Christina
        Martinez Jesse
        Pruitt Carolyn
        Swanson Michael
        Stewart Heather
        Rivera Kelly
        Allen Jamie
        Boone Jesse
        Weber Heather
        Nelson Paul
        Salazar Walter
        Delacruz Jason
        Davis Jennifer
        Fisher Mark
        Freeman John
        Mckenzie Michael
        Kelly Kelsey
        Soto Justin
        Berg Mr.Daniel
        Romero Paul
        Myers David
        Martin Laura
        Thompson Johnny
        Boyd Claudia
        Brown
        DDS Larry
        Olson Virginia
        Black Ryan
        Long Victoria
        Brown Susan
        Hogan Natalie
        Anderson Barbara
        Baker Sabrina
        Beck James
        Hubbard Theresa
        Duncan Kimberly
        Lopez Daniel
        Williams George
        Martin Natasha
        Waters Thomas
        Wheeler Nathaniel
        Gonzalez Ian
        White Adam
        Miller Carol
        Cooper Luis
        Wright Christian
        Ruiz Patrick
        Vega Patty
        Patel Mary
        Taylor Leah
        Collins
        DDS Joseph
        Solis Nathaniel
        Solomon Brianna
        Henderson Courtney
        Ramirez Karen
        Miller Amanda
        Barr Richard
        Edwards Laura
        Hammond Shirley
        Blackwell Angelica
        Kidd Nicole
        Wilson Shelby
        Herring Melissa
        Barnes Michael
        Davis Rebecca
        Michael Theresa
        Thompson Kristen
        Andrews Jamie
        Brown Danielle
        Tanner
        DDS Andrea
        Macias
        MD Charles
        Griffith Julie
        Gilbert Tracie
        Hartman Shelby
        Davies David
        Turner Kent
        Davis Michael
        Lawson Tonya
        Rodriguez Dr.Scott
        Gordon Jennifer
        Lyons Andrew
        Clark Ryan
        Hall Laura
        Ewing Christopher
        Wright Gordon
        Arnold Kelly
        Cooper Tammy
        Carter Kelly
        Garcia Scott
        Velazquez Tammy
        Roman Chelsea
        Barnett Alicia
        Alexander Jeremy
        Freeman Peter
        Andrade Margaret
        Cuevas James
        Baker Richard
        Horton Rachel
        Armstrong Brendan
        Fischer Amber
        Klein Anthony
        Lester Miguel
        Brown Tyler
        Washington Maria
        Joyce Michael
        Watts
        DDS Kim
        Watts Michael
        Jacobs Timothy
        Lopez Jennifer
        Gould Jennifer
        Gonzalez William
        Gomez Frank
        King Rachel
        Butler Paul
        Estrada Courtney
        Campbell Christine
        Lowe Edward
        Leonard John
        Sanchez Daniel
        Powell Amber
        Ortiz Pamela
        Simpson Kristopher
        Brown Michelle
        Hall Mr.Gregory
        Smith
        II Christopher
        Walsh Edward
        Ross Henry
        Vargas Nathan
        Evans Terry
        Brown Brian
        Davis Michele
        Sanchez Joshua
        Martinez Brenda
        Hernandez Kristina
        Wall Courtney
        White Brian
        Watts Donald
        Henderson Richard
        Davidson Lisa
        Stark Rachel
        Potter Christina
        Dodson Michele
        Gibson Gabrielle
        Brennan William
        Rivera, Matthew
        Fleming, Tara
        Copeland, Anthony
        Garrett, Megan
        Andrews, Mason
        Green, Kenneth
        Smith, Elijah
        Hughes, Amy
        Brown, Michael
        Hart, Christina
        Rogers, Jeremy
        Moody, Matthew
        Burton, Dennis
        Hoffman, Gregory
        Williams, George
        Santiago, Ryan
        Cox, Michele
        Adams, Michael
        Meadows, Gabrielle
        Mcdonald, Dr.Nicholas
        Schneider, Kelly
        Cline, Madison
        Richards, Nathan
        Barber, Marcus
        Fernandez, Daisy
        Barnett, Cody
        Johnson, Rick
        Allen, Steven
        Garcia, Robert
        Swanson, Catherine
        Robertson, Danielle
        Zamora, Christine
        Reid, Sarah
        Garcia, Lindsey
        Smith, Sally
        Baldwin, Joel
        Summers, Russell
        Prince, Robert
        Smith, Benjamin
        Taylor, Patricia
        Parker, Kristi
        Benson, Dakota
        Dorsey, Rhonda
        Smith, Nancy
        Brooks, Deanna
        Nelson, Cody
        York, Tiffany
        Owens, Patricia
        Hawkins, Laurie
        Grant, Jerry
        West, Brian
        Heath, Brittney
        Smith, Benjamin
        Parker, Kayla
        Harris, Lisa
        Tate, Lori
        Nunez, John
        Graham, Paul
        Nelson, Jennifer
        Green
        DVM, Felicia
        Diaz, Colleen
        Hernandez, Emma
        Sanchez, Susan
        Adams, Andrew
        Castillo, Patricia
        Barnett, Miss
        Katherine
        Davis
        DDS, Alicia
        Ellis, Nicholas
        Rodriguez, Jennifer
        Hunt, Julie
        Smith, Kendra
        Jimenez, Linda
        Frazier, Shelby
        Smith, James
        Scott, Patrick
        Robinson, Jerry
        Bell, Jason
        Payne, Jessica
        Foster, Jeffrey
        Moore, Linda
        Brown, James
        Rodriguez, Andre
        Williams, Joan
        Moore, Billy
        Conley, Margaret
        Mercado, Aaron
        Perry, Kristen
        Williams, Mary
        Henderson, Larry
        Garner, Charles
        Baxter, Catherine
        Wheeler, Daniel
        Alvarado, Kelly
        Reynolds, Gina
        Lutz, Shawn
        Boone, Mr.Adam
        Harris
        III, Lisa
        Davis, Robert
        Taylor, Jonathan
        Murray, Ruben
        Perez, Daniel
        Sloan, Carmen
        Delgado, Robert
        Ramos, Daniel
        Mcintyre, Edward
        Newman, Gabrielle
        Fleming, Robert
        Simmons
        MD, Amanda
        Moore, Kelly
        Robinson, James
        Acosta, Jill
        Rodriguez, Diane
        Robinson, Maria
        Oconnell, Heather
        Parker, Patricia
        Chan, Taylor
        Johnson, Michael
        Aguilar, Catherine
        Ramirez, Jerry
        Brown, Sierra
        Gutierrez, Michael
        Hudson, Marie
        Ramirez, James
        Williams, Daniel
        Thompson, Denise
        Johnson, Angel
        Espinoza, Krystal
        Archer, Christine
        Mccarthy, Mark
        Harris, Laura
        Bradley, Michelle
        Freeman, Keith
        Russell, Jesse
        Martinez
        DDS, Traci
        Green, Mrs.Angel
        Hamilton
        DDS, Laura
        Miller, Patrick
        Peterson, Adam
        Kaiser, Adam
        Kelly, Steven
        Pennington, John
        Irwin, Chelsea
        Maynard, Brittany
        Henderson, Courtney
        Clark, Cynthia
        Wall, Brooke
        Jones, Joshua
        Ramirez, Paul
        Nunez, Kevin
        Hall, Tara
        Hughes, Ethan
        Miles, Lisa
        Martin, Kevin
        Walker, Tamara
        Matthews, Kristin
        Carey, Sharon
        Meyers, Carrie
        Carpenter, Steve
        Long, Katie
        Young, Diana
        Barnes, Brian
        Riddle, Kristen
        Baker, Donald
        Wong, Brian
        Tran, Isabel
        Vazquez, Alexis
        Wood, Roy
        Robinson, Lisa
        Vasquez, Karen
        Shelton, Brian
        Kirk, Christy
        Koch, Adam
        Martinez, Kristi
        Vargas, Christopher
        Perez, Jennifer
        Dominguez, Jennifer
        Smith, Melissa
        Taylor, Sarah
        Hall, Kenneth
        Pope, Christina
        Williams, Mr.Andrew
        Banks, Kayla
        Ho, Jared
        Johnson, Melissa
        Kane, Bruce
        Thomas, Amy
        Lee, Juan
        Small, Gary
        Choi, Cynthia
        Armstrong, Dylan
        King, Robert
        Williams, Anthony
        Johnson, Andre
        Ross, Sara
        Martin, Brenda
        Mclean, Ralph
        Carroll, Michael
        Jones, Gregory
        Crawford, Christopher
        Wallace, Linda
        Velasquez, Stephen
        Perez, Rebecca
        Graves, Stephen
        Watson, Bradley
        Alexander, Jason
        Garner, Jacqueline
        Sanders, John
        Golden, Angela
        Owen
        MD, Andrew
        Walsh
        Jr., Rebecca
        Sanchez, Cameron
        Russo, Danielle
        Jennings, Mrs.Sarah
        Smith, William
        Gonzales, Amy
        Lewis, Timothy
        Davis, Nancy
        Shepard
        DVM, Stephanie
        Casey, Susan
        Ashley, Tonya
        Johnson, Jasmine
        Rodriguez, Raymond
        Hernandez, Kelly
        Porter, Teresa
        Walker, Sheila
        Gonzalez, Wanda
        Smith, Nancy
        Woodward, Amanda
        Wilson, Lisa
        Williams, Derek
        Robertson, Dustin
        Nguyen, Luis
        Klein, Amy
        Torres, Shelley
        White, Cody
        Bowers, Lisa
        Arnold, Tamara
        Evans, Whitney
        Boyer, Timothy
        Jones, Frank
        Houston, Kimberly
        Mcclure, Justin
        Gibson, Gregory
        Davidson, Christopher
        Stone, Sherry
        Rodriguez, Lacey
        Guerra, Timothy
        Gilbert, Joan
        Munoz, Morgan
        Smith, Troy
        Wallace, Andrew
        Schmidt, Tammy
        Spencer, Melissa
        Holloway, Monica
        Jones, Dominique
        Collier, Francisco
        Rogers, Michelle
        Marsh, William
        Morris, Matthew
        Reed, Suzanne
        Martin, Valerie
        Flores, Kathleen
        Garcia, Anna
        Mills, Mckenzie
        Thomas, Teresa
        Jenkins, Carol
        Lyons, Victor
        Marshall, Kimberly
        Carpenter, Glenn
        Costa, Robin
        Larson, Jessica
        Ford, Briana
        Holden, Rachael
        Williams, Ms.Carol
        Schroeder, James
        Estes, Lauren
        Williams, Andrew
        Hopkins, Tyler
        Estrada, David
        Velez, Henry
        Ramos, Mark
        White, John
        Flowers, Robert
        Flores, Pamela
        Finley, Michelle
        Watkins, Logan
        Carter, Daniel
        Diaz, Charles
        Deleon, Shawn
        Parker, Anthony
        Fuller, Andrew
        Hayden, Jessica
        Velasquez, John
        Hutchinson, Robin
        Davis, Maria
        Spence, Angela
        Sullivan, Margaret
        Arias, Robert
        Bennett, David
        Scott, Jennifer
        Perez, David
        Harrington, Becky
        Thomas, Tanya
        Hall, Christopher
        Ray, Mark
        Mason, Lindsay
        Cain
        MD, Susan
        Patterson, Aaron
        Roberts, Michael
        Myers, Diane
        Garcia, Jeffrey
        Hammond, Christopher
        Williams, Yvonne
        Jordan, Carolyn
        Lewis, Elizabeth
        Gonzalez, Shannon
        Villarreal, Mr.Matthew
        Smith
        DVM, Shannon
        Phillips, Donald
        Mccoy, Mallory
        Ali, Shawn
        Allen, Traci
        Smith, Paige
        Patterson, Rachael
        Young, Michael
        Flynn, Bryan
        Sanders, Brittany
        Spencer, Andrew
        Miller, Ryan
        Fowler, Meagan
        Simon
        DDS, Michael
        Hoffman, Joshua
        Haas, Karen
        Anthony, Kenneth
        Davidson, Zachary
        Randolph, Brian
        Rivers, Jared
        King, Robert
        Jones, James
        Shepherd, Matthew
        Lopez, Tracey
        Bennett, Lori
        Weber, Devin
        Garza, Robert
        Liu, Carla
        Bartlett, Michael
        Carson, Mr.Andrew
        Phillips, Dr.Charles
        Torres
        DDS, Sara
        Bowen, Kristin
        Guerrero
        MD, Elizabeth
        Austin, Elizabeth
        Adams, Benjamin
        Jones, Victoria
        Miller, Jennifer
        Morgan
        MD, Gregory
        Jones, Timothy
        Schneider, Alex
        Prince, Charles
        Bell, Amy
        Garcia, Michael
        Middleton, Raymond
        Johnson, Dawn
        Bartlett, Theodore
        Oconnor, Sarah
        Garcia, Sean
        Vega, Victoria
        Baker, Jamie
        Wilson, Michelle
        Skinner, Alexa
        Nguyen, Dennis
        Franklin, Mr.Ian
        Patel, Danielle
        Cowan, Jason
        May, Andrea
        Hernandez, Latasha
        Dixon, Christine
        Soto, Paul
        Lloyd, David
        Bruce, Mark
        Ramsey, Gail
        Rodriguez, Lisa
        Galvan, Wanda
        Watson, Dr.Stephanie
        Nguyen, Elizabeth
        Watts, John
        Hensley, Rebecca
        Silva, Kelly
        Beltran, Elizabeth
        Anthony, Raymond
        Ortiz, Joseph
        Salinas, Regina
        Anderson, Leonard
        Tucker, Jose
        Hubbard, Katherine
        Nichols, Jerry
        Wong, Tyler
        Wright, Taylor
        Anderson, Joshua
        Young, Alexander
        Duncan, Robert
        Lopez, Theresa
        Morris, Daniel
        Miller, Michael
        Neal, Tracy
        Dougherty, Dr.Jennifer
        Morales, Kristopher
        Stout, Julie
        Fuller, Kristen
        Cowan, Jesse
        Knight, Alexander
        Ibarra, David
        Lowe, Rebecca
        Sharp, Mr.Thomas
        King, Steven
        Wiley, Elizabeth
        Davis, Brett
        Ferrell, Brandon
        Stanley, Christina
        Lee, James
        Weeks, Timothy
        Wade, David
        Smith, Andrew
        Preston, William
        Rodriguez, Valerie
        Jones, Cheryl
        Sanchez, Marco
        Scott, Stephen
        Gould, Diana
        Reynolds, Jillian
        Lamb, Scott
        Fisher, Nicholas
        Mcintyre, Jose
        Mayer, Monique
        Garza, Steven
        Greer, April
        Pearson, Maria
        Everett, Denise
        Fletcher, Kelly
        Mccarthy, Cameron
        Benson, David
        Riley, Maria
        Franklin, Erica
        Harris, Caitlin
        Grant
        DDS, Pamela
        Walker, Linda
        Hobbs, Katherine
        Haas, Angelica
        Johnson, Derek
        Harrington, Mary
        Garcia, Amanda
        Walter, Rachel
        Stewart, Angel
        White, Michael
        Jones, Vanessa
        Carpenter, Amanda
        Ford, Brittany
        Dawson, Morgan
        Fox, Robin
        Gonzales, Anthony
        Hansen, Jennifer
        Figueroa, Timothy
        Morales, Bernard
        Roberts, David
        Alexander, Sarah
        Weiss, Amy
        Garcia, Kathryn
        Harris, Andrea
        Bennett, Amy
        Henry, Juan
        Hart, Jennifer
        Hines, Jessica
        Cunningham, Justin
        Brown, Kenneth
        Wilson, Christopher
        Reed, Kirsten
        Harmon, William
        Jones, Mark
        Davis, Teresa
        Johnson, Anthony
        Key, Maureen
        Jones, Richard
        Noble, Eric
        Smith, Scott
        Haynes, David
        Bishop, Tina
        Aguilar, Dr.Elizabeth
        Lopez, Lisa
        Hill, Michael
        Flores, Christopher
        Green, Gregory
        Elliott, Carlos
        Berg
        DVM, Holly
        Gonzalez, Ryan
        Young, Nathaniel
        Peterson, Barbara
        Miller, Theresa
        Chase, Valerie
        Murphy, Richard
        Ramirez, Brett
        Yates, James
        Huff, Richard
        Maxwell, Kenneth
        Pace, Matthew
        York, Vanessa
        Davidson, Courtney
        Berg, Steven
        Ayala, Daniel
        Osborne, Justin
        Daniels, Meghan
        Jacobs, Michelle
        Mendez, Katherine
        Hardy, Jamie
        Knox, James
        Washington, Mariah
        Patel, Scott
        Brown, Shane
        Mills, Angie
        Jennings, Desiree
        Goodwin, Gregory
        Lewis, Lauren
        Carney, Jonathan
        Martin, Dr.Kathryn
        Watson
        MD, Jimmy
        Moore, Devon
        Cole, Mary
        Howard, Diane
        Henderson, Kirk
        Baker, Felicia
        Wallace, Paul
        Smith, Heather
        Hampton, David
        Castaneda, Shannon
        Jackson, Dr.Ashley
        Yang, Kelly
        Nelson, Samantha
        Cortez, Jennifer
        Santos, Stephen
        Green, Kim
        Kaufman, Douglas
        Smith, Paul
        Nixon, Abigail
        Price, John
        Murillo, Dr.Joseph
        Lee, Joseph
        Daniel, Christina
        Morales, Christopher
        Schmidt, Christopher
        Wang, Joseph
        Morris, Holly
        Meyers, Craig
        Kelley, Anthony
        Jefferson, Ellen
        Smith, Travis
        Garcia
        Jr., Shannon
        Weaver, Carla
        Clayton, Jessica
        Escobar, Michael
        Russell, Vanessa
        Pena, Jimmy
        Edwards, William
        Wilkerson, Peter
        Berger, Sarah
        Daniel, Zachary
        Burnett, Mr.Calvin
        Anderson, Courtney
        Wade, Frank
        Lambert, Kimberly
        Smith, Melanie
        Miller, Shawn
        Ramirez, Christina
        Solomon, Eric
        Moore, Joshua
        Velazquez, Tyler
        Graham, Cameron
        Wells, Brian
        Brown, Juan
        Barnett, Elizabeth
        Matthews, Brandon
        Brown, Emily
        Hunter, Erika
        Gregory, Craig
        Roberts, Jeffrey
        Smith, Angela
        Ward, Brandi
        Allen, Joseph
        Baxter, Michael
        Clark, Candice
        Singleton, Claudia
        Wang, Amber
        Perez, Michael
        Brown, Melissa
        Campbell, Vanessa
        Hernandez, Andrea
        Wilson, Nathaniel
        Petersen, Jesse
        Leon, Karen
        Martin, Dr.Kenneth
        Jones, Timothy
        Gonzalez, Mrs.Sharon
        Brown, Reginald
        Larson, James
        Jordan, Alexis
        Novak, Jessica
        Foster, Daniel
        Black, Elizabeth
        Martin, Gregory
        Nguyen
        DVM, Nicholas
        Robinson, Alejandra
        Mclaughlin, Mrs.Victoria
        Ortiz
        DDS, Bradley
        Greene, Mary
        White, Shannon
        Meyer, Joshua
        Campos, Samantha
        Henderson, Jackson
        Wall, Molly
        Weber, Rebecca
        Olson, Elizabeth
        Fitzpatrick, Latasha
        Wilson, Jennifer
        Rivera, Mrs.Haley
        Davis, Gregory
        Knight, Heather
        Nelson, Sherry
        Lawson, Debbie
        Woodward, Mark
        Farrell, Eric
        Gross
        DVM, Maria
        Liu, Darryl
        Mckay, James
        Lucas, Marcus
        Graves, Robert
        Castro, Nicole
        Goodwin, Sheri
        Koch, Keith
        Livingston, Nathan
        Henry, Jose
        Ruiz, Kurt
        Haas, Shane
        Anderson, Jason
        Gray, Kathy
        Owen, Brian
        Williams, Valerie
        Peters, Cameron
        Guzman, Frank
        Roberson, David
        Garrett, Ryan
        Wallace, Tiffany
        Brown, Stephen
        Lopez, Charles
        Scott, Erin
        Newman, Steven
        Wade, Scott
        Valenzuela, Dustin
        Williams, Vicki
        Harrison, Jessica
        Calderon, Alexis
        Gonzales, Randy
        Henderson, Catherine
        Navarro, Gabriella
        Alvarez, Kimberly
        Burton, Paul
        Bailey, Jaime
        Turner, Kimberly
        White, Christine
        Campbell, Thomas
        Spencer, Martha
        Baker, James
        Butler, George
        Reynolds, Victoria
        Patrick, Donna
        Sutton, Paul
        Nicholson, Mark
        Juarez, Sarah
        Carson, Jonathan
        Robinson, Judy
        Hobbs, Nancy
        Collins, Megan
        Jones, Amanda
        Reid, Victor
        Blackwell, Justin
        Ruiz, Ryan
        Stafford, Ryan
        Clark, Christina
        Carlson, Laura
        Ramirez, David
        Delacruz, Sara
        George, Jamie
        Gonzalez, Elizabeth
        Rios, Willie
        Aguilar, Austin
        Sanchez, Kyle
        Carey, Kelly
        Duran, Andrew
        Valentine, Paul
        Williamson, Martin
        Knight, Kelly
        Butler, Emily
        Middleton, Michael
        Young, Jimmy
        Sloan, Sherry
        Gibson, Jose
        Turner, James
        Jones, Elizabeth
        Scott, Nancy
        Morris, Brian
        Tran, Rebecca
        Casey, Susan
        Buchanan, Jeffrey
        Kane, Betty
        Flynn, Brooke
        Davis, Sarah
        Powell, James
        Hudson, John
        Santiago, Mckenzie
        Collier, Taylor
        Marquez, Trevor
        Jones, James
        Morris, Mr.Thomas
        Elliott, Elizabeth
        Bartlett, Sandra
        Buckley, Diane
        Castro, Mercedes
        Ross, Jessica
        Pacheco, Katie
        Waters, Gregory
        Medina, Michelle
        Gray, Shari
        Carter, Selena
        Ross, Jeremy
        Miller, Jordan
        York, Ashley
        Figueroa, Julie
        Arnold, Sarah
        Willis, Laura
        Dorsey, Sandra
        Flores, Linda
        Coleman, Jean
        Jones, Michael
        Valdez, Kristin
        Weaver, Jessica
        Rivera, Erik
        Peterson, Richard
        Smith, Gabriela
        Dean
        DDS, Tracy
        Bowen, Timothy
        Gray, Jillian
        Pearson, Barbara
        Jackson, Chase
        Barton, Joel
        Hood, Jose
        Cruz, Brian
        Whitaker, Patricia
        Anderson, Kimberly
        Boyd, Anthony
        Ortiz, Sherry
        Hunter, Catherine
        Cohen, Rhonda
        Perez, Patricia
        Morris, Robert
        Johnson, Autumn
        Gallagher, Eric
        Clements
        """.splitlines(True) if name.strip()][:limit]



import random
import uuid
import datetime
import tzlocal
import psutil
from .generator import Generators as g
import time
from threading import Thread, Event
from queue import Queue
import statistics


class KawaBenchmark:

    def __init__(self,
                 kawa_client,
                 metrics_output_file,
                 view_ids,
                 user_ids,
                 num_users,
                 benchmark_name='Benchmark',
                 num_seconds_between_two_computations=1,
                 test_duration_in_seconds=60):

        """
        Creates a new KAWA benchmark object.
        In order to run it, just call benchmark.run()

        :param kawa_client: An instance of the kawa client,
                            connected to the workspace in which you want to run your tests.
        :param benchmark_name: The name of the benchmark will appear in the output csv file.
        :param metrics_output_file: Path to the benchmark output file. It will be a csv.
        :param view_ids: The pool of view to compute during the benchmark.
        :param user_ids: The pool of users that will be used to compute the views.
        :param num_users: The total number of users that will be simulated in the benchmark
        :param num_seconds_between_two_computations:
        :param test_duration_in_seconds:
        """

        self._k = kawa_client
        self._benchmark_name = benchmark_name
        self._metrics_output_file = metrics_output_file
        self._view_ids = [str(v) for v in view_ids]
        self._user_ids = [str(v) for v in user_ids]
        self._num_users = num_users
        self._num_seconds_between_two_computations = num_seconds_between_two_computations
        self._test_duration_in_seconds = test_duration_in_seconds
        self._cache_hit_ratio = 35

        # Reporting
        self._computation_times_per_user = {}
        for user_id in self._user_ids:
            self._computation_times_per_user[user_id] = []

        self._computation_times_per_view = {}
        for view_id in self._view_ids:
            self._computation_times_per_view[view_id] = []

        # Synchronization
        self._end_of_test_event = Event()
        self._test_timing_queue = Queue()

    def run(self):
        computation_threads = []
        metrics_thread = Thread(target=self._system_metrics_thread)
        try:
            for thread_id in range(0, self._nb_threads()):
                thread = Thread(target=self._computing_thread)
                computation_threads.append(thread)
            [t.start() for t in computation_threads]
            metrics_thread.start()
            [t.join() for t in computation_threads]
        finally:
            self._end_of_test_event.set()
            if metrics_thread.is_alive():
                metrics_thread.join()
            self._print_results()

    def _print_results(self):
        all_compute_times = []

        print('--- STATS PER VIEW')
        for view_id, computation_times in self._computation_times_per_view.items():
            all_compute_times.extend(computation_times)
            self.extract_stats(timings=computation_times, item_to_print='View {}'.format(view_id))

        print('--- STATS PER USER')
        for user_id, computation_times in self._computation_times_per_user.items():
            self.extract_stats(timings=computation_times, item_to_print='User {}'.format(user_id))

        print('--- OVERALL STATS')
        self.extract_stats(timings=all_compute_times, item_to_print='Overall')

    def _nb_threads(self):
        # Each thread will start one computation every second
        # Hence: The number of threads is equal to the number of computations per second
        num_computations_per_second = int(self._num_users / self._num_seconds_between_two_computations)

        if num_computations_per_second < 1:
            raise Exception('The computation frequency is too low, please ensure at least 1 computation/s')

        return num_computations_per_second

    def _random_filter(self):
        random_between_1_and_100 = random.randint(1, 100)
        cache_hit = random_between_1_and_100 < self._cache_hit_ratio
        return 'וָו' if cache_hit else 'וָו' + str(uuid.uuid4())

    def _compute_one_view(self, view_id, user_id):
        lazy_frame = (self._k.sheet(no_output=True)
                      .view_id(view_id=view_id)
                      .as_user_id(as_user_id=user_id)
                      .filter(self._k.col('{--}{=æ=}{--}').does_not_start_with(self._random_filter()))
                      .limit(300))
        start = time.time()
        lazy_frame.compute()
        computation_time = time.time() - start

        self._computation_times_per_user[user_id].append(computation_time)
        self._computation_times_per_view[view_id].append(computation_time)
        self._test_timing_queue.put_nowait(computation_time)

    def _computing_thread(self):
        thread_start = time.time()
        while True:

            random_sleep = random.random()
            time.sleep(random_sleep)

            view_id = g.get_random_element_in(self._view_ids)
            user_id = g.get_random_element_in(self._user_ids)
            self._compute_one_view(view_id, user_id)

            time.sleep(1.0 - random_sleep)

            thread_seconds_elapsed = time.time() - thread_start
            if thread_seconds_elapsed > self._test_duration_in_seconds:
                break

    def _system_metrics_thread(self):
        print('Dumping system metrics in {}'.format(self._metrics_output_file))
        with open(self._metrics_output_file, 'w') as f:
            counter = 0
            header = 'TIME,TEST NAME,CPU %,MEM %,NB READ,NB WRITE,NB BYTES READ,NB BYTES WRITE,NUM TESTS,MIN,MAX,AVG'
            f.write(header + '\n')
            print(header)

            initial_io = psutil.disk_io_counters(perdisk=False)
            prev_io = initial_io

            while not self._end_of_test_event.is_set():

                test_timings = []
                while not self._test_timing_queue.empty():
                    test_timings.append(self._test_timing_queue.get(block=False))

                time.sleep(1)
                counter = counter + 1

                dt = datetime.datetime.now(tz=tzlocal.get_localzone())
                cpu_percent = psutil.cpu_percent(percpu=False)
                virtual_mem = psutil.virtual_memory()
                mem_percent = virtual_mem.percent

                metrics = [dt.strftime('%Y-%m-%d %H:%M:%S'), self._benchmark_name, cpu_percent, mem_percent]

                current_io = psutil.disk_io_counters(perdisk=False)
                delta_read_count = int(current_io.read_count - prev_io.read_count)
                delta_write_count = int(current_io.write_count - prev_io.write_count)
                delta_read_bytes = int(current_io.read_bytes - prev_io.read_bytes)
                delta_write_bytes = int(current_io.write_bytes - prev_io.write_bytes)

                metrics.extend([delta_read_count, delta_write_count, delta_read_bytes, delta_write_bytes])
                metrics.extend(self.extract_stats(test_timings))

                line = ','.join([str(m) for m in metrics])

                print(line)
                f.write(line + '\n')

                prev_io = current_io

    @staticmethod
    def extract_stats(timings, item_to_print=None):
        if not timings:
            s = [0, 0.0, 0.0, 0.0]
        else:
            average = round(statistics.mean(timings), 2)
            minimum = round(min(timings), 2)
            maximum = round(max(timings), 2)
            s = [len(timings), minimum, maximum, average]

        if item_to_print:
            print("[{}] {} computations: MIN={} MAX={} AVG={}".format(item_to_print, s[0], s[1], s[2], s[3]))

        return s
